from __future__ import annotations
from dataclasses import dataclass
import typing

from aeon.core.liquid import LiquidApp, LiquidLiteralFloat
from aeon.core.liquid import LiquidLiteralBool
from aeon.core.liquid import LiquidLiteralInt
from aeon.core.liquid import LiquidLiteralString
from aeon.core.liquid import LiquidTerm
from aeon.core.liquid import LiquidVar
from aeon.core.types import (
    AbstractionType,
    LiquidHornApplication,
    RefinedType,
    Top,
    Type,
    TypeConstructor,
    TypePolymorphism,
    TypeVar,
    t_bool,
)
from aeon.core.types import t_float
from aeon.core.types import t_int
from aeon.core.types import t_unit
from aeon.core.types import t_string
from aeon.prelude.prelude import native_types
from aeon.typechecking.context import (
    TypeBinder,
    TypeConstructorBinder,
    TypingContext,
    UninterpretedBinder,
    VariableBinder,
)

from aeon.utils.name import Name


class LiquidTypeCheckException(Exception):
    pass


@dataclass
class LiquidTypeCheckingContext:
    known_types: list[TypeConstructor]
    variables: dict[Name, TypeConstructor | TypeVar]
    functions: dict[Name, list[TypeConstructor | TypeVar]]

    def __repr__(self):
        kt = "; ".join([str(t) for t in self.known_types])
        vars = "; ".join([str(t) + ":" + str(self.variables[t]) for t in self.variables])
        fns = "; ".join([str(t) for t in self.functions])
        return f"(LiquidGamma {kt} | {vars} | {fns} )"


def lower_abstraction_type(ty: Type) -> list[TypeConstructor | TypeVar]:
    args: list[TypeConstructor | TypeVar] = []
    while True:
        match ty:
            # TODO: Should these be removed?
            case Top() | RefinedType(_, Top(), _):
                return args + [t_unit]
            case TypeVar(_):
                assert args
                return args + [ty]
            case (
                AbstractionType(_, RefinedType(_, aty, _), RefinedType(_, rty, _))
                | AbstractionType(_, RefinedType(_, aty, _), rty)
                | AbstractionType(_, aty, rty)
            ):
                match aty:
                    case TypeConstructor(_, _) | TypeVar(_):
                        args.append(aty)
                    case Top():
                        args.append(t_unit)
                    case _:
                        # For Higher-order functions, we use an int parameter.
                        args.append(t_int)
                ty = rty
            case TypePolymorphism(_, _, body):
                return lower_abstraction_type(body)
            case TypeConstructor(_, _):
                return args + [ty]
            case RefinedType(_, bt, _):
                return args + [bt]
            case _:
                assert False, f"Unknown type {ty} when lowering to liquid"


T = typing.TypeVar("T")


def flatten(xs: list[list[T]]) -> list[T]:
    return [x for y in xs for x in y]


def lower_context(ctx: TypingContext) -> LiquidTypeCheckingContext:
    known_types: list[Name] = native_types + []
    variables: dict[Name, TypeConstructor | TypeVar] = {}
    functions = {}

    for entry in ctx.entries[::-1]:
        match entry:
            case VariableBinder(name, TypeConstructor(_) as bt):
                variables[name] = bt
            case VariableBinder(name, TypeConstructor(_, _) as bt):
                variables[name] = bt
            case VariableBinder(name, TypeVar(tvname)):
                known_types.append(tvname)
                variables[name] = TypeVar(tvname)
            case VariableBinder(name, TypePolymorphism(_, _, _) as ty):
                functions[name] = lower_abstraction_type(ty)
            case TypeBinder(name, _):
                known_types.append(name)
            case UninterpretedBinder(name, AbstractionType(_, _, _) as ty) | VariableBinder(
                name, AbstractionType(_, _, _) as ty
            ):
                functions[name] = lower_abstraction_type(ty)
            case UninterpretedBinder(name, TypePolymorphism(_, _, AbstractionType(_, _, _)) as ty):
                functions[name] = lower_abstraction_type(ty)
            case VariableBinder(name, RefinedType(_, TypeConstructor(_) as bt, _)):
                variables[name] = bt
            case VariableBinder(name, RefinedType(_, TypeConstructor(_, _) as bt, _)):
                variables[name] = bt
            case TypeConstructorBinder(_, _):
                pass
            case _:
                assert False, f"Unknown context type {entry} ({type(entry)})"

    return LiquidTypeCheckingContext([TypeConstructor(n) for n in known_types], variables, functions)


def type_infer_liquid(
    ctx: LiquidTypeCheckingContext,
    liq: LiquidTerm,
) -> TypeConstructor | TypeVar:
    match liq:
        case LiquidLiteralBool(_):
            return t_bool
        case LiquidLiteralInt(_):
            return t_int
        case LiquidLiteralFloat(_):
            return t_float
        case LiquidLiteralString(_):
            return t_string
        case LiquidVar(name):
            if name not in ctx.variables:
                raise LiquidTypeCheckException(f"Variable {name} not in context in {ctx}.")

            match ctx.variables[name]:
                case TypeConstructor(_, []) as ty:
                    return ty
                case TypeConstructor(_, _) as ty:
                    return ty
                case TypeVar(_) as ty:
                    return ty
                case _:
                    raise LiquidTypeCheckException("Could not find type for {liq} ({ctx.variables[name]})")
        case LiquidApp(fun, args):
            if fun not in ctx.functions:
                raise LiquidTypeCheckException(f"Function {fun} not in context in {liq} ({ctx.functions}).")
            ftype = ctx.functions[fun]
            equalities: dict[Name, TypeConstructor | TypeVar] = {}

            def resolve_type(ty: Type) -> TypeConstructor | TypeVar:
                match ty:
                    case TypeConstructor(_, _):
                        return ty
                    case TypeVar(tv):
                        if tv in equalities:
                            return resolve_type(equalities[tv])
                        else:
                            return ty
                    case _:
                        assert False, "unknown stuff"

            if len(ftype) != len(args) + 1:
                raise LiquidTypeCheckException(
                    f"Function application {liq} needs {len(ftype) - 1} arguments, but was passed {len(args)}."
                )
            type_of_args = []

            for arg, exp_t in zip(args, ftype):
                k = type_infer_liquid(ctx, arg)
                type_of_args.append(k)
                match (k, exp_t):
                    case (TypeConstructor(t), TypeConstructor(e)):
                        if t != e:
                            raise LiquidTypeCheckException(
                                f"Argument {arg} in {liq} is expected to be of type {exp_t}, but {k} was found instead."
                            )
                    case (TypeConstructor(t), TypeVar(name)):
                        if name not in equalities:
                            equalities[name] = resolve_type(k)
                        elif resolve_type(TypeVar(name)) != k:
                            raise LiquidTypeCheckException(
                                f"Argument {arg} in {liq} is expected to be of type {exp_t} ({equalities[name]}), but {k} was found instead."
                            )
                    case (TypeVar(t), TypeVar(name)):
                        if name not in equalities:
                            equalities[name] = resolve_type(k)
                        elif resolve_type(TypeVar(name)) != k:
                            raise LiquidTypeCheckException(
                                f"Argument {arg} in {liq} is expected to be of type {exp_t}, but {k} was found instead."
                            )
                    case _:
                        raise LiquidTypeCheckException(
                            f"Could not unify {k} ({type(k)}) and {exp_t} ({type(exp_t)}) in {liq}."
                        )

            def is_base_type_in(t: Type, names: list[str]) -> bool:
                match t:
                    case TypeConstructor(Name(name, _)):
                        return name in names
                    case _:
                        return False

            first_argument = type_of_args[0]
            match fun:
                case Name(fun_name, _):
                    if fun_name in ["<", "<=", ">", ">="]:
                        if not is_base_type_in(first_argument, ["Float", "Int"]) and not isinstance(
                            type_of_args[0], TypeVar
                        ):
                            raise LiquidTypeCheckException(f"Function {fun_name} only applies to Floats or Ints.")
                    elif fun_name in ["==", "!="] and not isinstance(first_argument, TypeVar):
                        # TODO: Add type equality
                        if not is_base_type_in(first_argument, ["Unit", "Bool", "Float", "Int", "String"]):
                            raise LiquidTypeCheckException(f"Function {fun_name} only applies to built-in types.")
                    elif fun_name in ["+", "-", "*", "/"] and not isinstance(first_argument, TypeVar):
                        if not is_base_type_in(first_argument, ["Float", "Int"]):
                            raise LiquidTypeCheckException(f"Function {fun_name} only applies to Floats or Ints.")
            return resolve_type(ftype[-1])
        case LiquidHornApplication(_, _):
            return t_bool
        case _:
            assert False, f"Constructed {liq} ({type(liq)}) not supported."


def check_liquid(
    ctx: LiquidTypeCheckingContext,
    liq: LiquidTerm,
    exp: TypeConstructor,
) -> bool:
    try:
        t = type_infer_liquid(ctx, liq)
        return t == exp
    except LiquidTypeCheckException:
        return False


def typecheck_liquid(
    ctx: TypingContext,
    liq: LiquidTerm,
) -> Type | None:
    assert isinstance(ctx, TypingContext)
    v = type_infer_liquid(lower_context(ctx), liq)
    return v



from aeon.decorators import decorators_environment
from aeon.sugar.program import Definition, SRec, STerm, SVar
from aeon.sugar.program import Decorator
from aeon.sugar.stypes import STypeConstructor, SType
from aeon.sugar.ast_helpers import st_float
from aeon.utils.name import Name, fresh_counter

single_objective_decorators = ["minimize", "maximize", "assert_property"]
multi_objective_decorators = [
    "multi_minimize",
    "multi_maximize",
    "assert_properties",
]


def get_minimize(minimize: list[bool]):
    if len(minimize) == 1:
        return minimize[0]
    else:
        return minimize


def get_type_from_decorators(macro_list) -> STypeConstructor:
    if len(macro_list) == 1:
        if macro_list[0].name in single_objective_decorators:
            return st_float
        elif macro_list[0].name in multi_objective_decorators:
            # TODO: replace -1 with the index of List, if defined as a special type
            # Or array.
            return STypeConstructor(Name("List", -1))
        else:
            raise Exception(
                "decorator not in lists single and multi objective decorators",
            )
    else:
        raise NotImplementedError("Not yet supported")


def extract_fitness_from_synth(d: Definition) -> tuple[STerm, list[Decorator]]:
    decorators_list: list[Decorator] = d.decorators
    assert len(decorators_list) > 0

    fitness_terms: list[STerm] = []
    for dec in decorators_list:
        annotation_func = getattr(decorators_environment, str(dec.name))
        expr_term = annotation_func(dec.macro_args)
        add_to_list(expr_term, fitness_terms)

    assert len(fitness_terms) > 0

    fitness_return_type = get_type_from_decorators(decorators_list)

    fitness_function = generate_term(
        d.name,
        fitness_return_type,
        fitness_terms,
    )

    return fitness_function, decorators_list


def add_to_list(item: list, my_list: list):
    try:
        my_list += item if isinstance(item, list) else [item]
    except TypeError as e:
        raise TypeError(f"An error occurred while adding to the list: {e}")

    return my_list


def generate_definition(
    fitness_args: list[tuple[str, SType]],
    fitness_return_type: STypeConstructor,
    fitness_terms: list[STerm],
) -> Definition:
    if len(fitness_terms) == 1:
        return Definition(
            name=Name("fitness", fresh_counter.fresh()),
            foralls=[],
            args=[],
            type=fitness_return_type,
            body=fitness_terms[0],
        )
    else:
        raise NotImplementedError("Not yet supported")


def generate_term(
    fitness_name: Name,
    fitness_return_type: STypeConstructor,
    fitness_terms: list[STerm],
) -> STerm:
    if len(fitness_terms) == 1:
        rec_name = Name(f"fitness_{fitness_name}", fresh_counter.fresh())
        return SRec(
            var_name=rec_name,
            var_type=fitness_return_type,
            var_value=fitness_terms[0],
            body=SVar(rec_name),
        )
    else:
        raise NotImplementedError("Not yet supported")



from __future__ import annotations

from abc import ABC
from dataclasses import make_dataclass
from typing import Generator, Optional, Tuple, Protocol
from typing import Type as TypingType

from geneticengine.grammar import extract_grammar, Grammar
from geneticengine.grammar.decorators import weight
from geneticengine.grammar.metahandlers.base import MetaHandlerGenerator
from geneticengine.prelude import abstract

from aeon.core.liquid import (
    LiquidApp,
    LiquidHole,
    LiquidLiteralBool,
    LiquidLiteralFloat,
    LiquidLiteralInt,
    LiquidLiteralString,
    LiquidTerm,
    LiquidVar,
)
from aeon.core.substitutions import substitution_in_type, substitution_liquid_in_type
from aeon.core.terms import Abstraction, Annotation, Application, If, Literal
from aeon.core.terms import Var
from aeon.core.types import AbstractionType, Type, TypePolymorphism, TypeVar
from aeon.core.types import TypeConstructor
from aeon.core.types import RefinedType
from aeon.core.types import Top
from aeon.core.types import t_bool
from aeon.core.types import t_float
from aeon.core.types import t_int
from aeon.core.types import t_string
from aeon.decorators import Metadata
from aeon.synthesis.grammar.mangling import mangle_name, mangle_var, mangle_type
from aeon.synthesis.grammar.refinements import refined_type_to_metahandler
from aeon.synthesis.grammar.utils import prelude_ops, aeon_to_python
from aeon.typechecking.context import (
    TypeBinder,
    TypeConstructorBinder,
    TypingContext,
    TypingContextEntry,
    UninterpretedBinder,
    VariableBinder,
)
from aeon.core.liquid_ops import ops
from aeon.utils.name import Name, fresh_counter


VAR_WEIGHT = 10000
LITERAL_WEIGHT = 3000
IF_WEIGHT = 1
APP_WEIGHT = 1000
ABS_WEIGHT = 100


def extract_class_name(class_name: str) -> str:
    prefixes = [
        "var_",
        "app_",
        "refined_app_",
        "refined_var_",
        "literal_Refined_",
        "literal_",
    ]
    for prefix in prefixes:
        if class_name.startswith(prefix):
            return class_name[len(prefix) :]
    return class_name


class GrammarError(Exception):
    pass


# Protocol for classes that can have a get_core method
class HasGetCore(Protocol):
    def get_core(self): ...


classType = TypingType[HasGetCore]


def is_valid_class_name(class_name: str) -> bool:
    return class_name not in prelude_ops and not class_name.startswith(
        ("_anf_", "target"),
    )


ae_top = type("ae_top", (ABC,), {})


def extract_all_types(types: list[Type]) -> dict[Type, TypingType]:
    data: dict[Type, TypingType] = {Top(): ae_top}
    for ty in set(types):
        if ty not in data:
            match ty:
                case TypeConstructor(_, args):
                    assert not args, "Polytypes in not supported in Synthesis"
                    class_name = mangle_type(ty)
                    ty_abstract_class = type(
                        class_name,
                        (ae_top, ABC),
                        {},
                    )
                    # ty_abstract_class = abstract(ty_abstract_class)
                    ty_abstract_class = abstract(ty_abstract_class)
                    data[ty] = ty_abstract_class
                case RefinedType(_, itype, _):
                    class_name = mangle_type(ty)
                    data.update(extract_all_types([itype]))
                    parent = data[itype]
                    ty_abstract_class = type(class_name, (parent, ABC), {})
                    ty_abstract_class = abstract(ty_abstract_class)
                    data[ty] = ty_abstract_class
                    # TODO: alpha-equivalence
                    # TODO: subtyping
                case AbstractionType(var_name, var_type, return_type):
                    class_name = mangle_type(ty)
                    data.update(
                        extract_all_types(
                            [var_type, substitution_in_type(return_type, Var(Name("__self__", 0)), var_name)]
                        )
                    )
                    ty_abstract_class = type(class_name, (ae_top, ABC), {})
                    ty_abstract_class = abstract(ty_abstract_class)
                    data[ty] = ty_abstract_class
                    # TODO: alpha-equivalence
                case Top():
                    data[ty] = ae_top
                case TypePolymorphism(_, _, _):
                    # TODO: Polytypes in Synthesis
                    continue
                case _:
                    assert False, f"Unsupported {ty}"
    return data


def create_literal_class(
    aeon_type: Type, parent_class: type, value_type: None | type | MetaHandlerGenerator = None
) -> type:
    """Create and return a new literal class with the given name and value type, based on the provided abstract class."""
    if value_type is None:
        value_type = aeon_to_python[aeon_type]

    class_name = mangle_type(aeon_type)
    new_class = make_dataclass(
        f"literal_{class_name}",
        [("value", value_type)],
        bases=(parent_class,),
    )

    def get_core(self):
        value = getattr(self, "value", None)
        return Literal(value, type=aeon_type)

    setattr(new_class, "get_core", get_core)
    new_class = weight(LITERAL_WEIGHT)(new_class)
    return new_class


def create_literals_nodes(type_info: dict[Type, TypingType], types: Optional[list[Type]] = None) -> list[TypingType]:
    """Creates all literal nodes for known types with literals (bool, int, float, string)"""
    if types is None:
        types = [t_bool, t_int, t_float, t_string]

    xname = Name("x", fresh_counter.fresh())

    gtm1 = LiquidApp(Name("<=", 0), [LiquidLiteralInt(-1), LiquidVar(xname)])
    lt256 = LiquidApp(Name("<=", 0), [LiquidVar(xname), LiquidLiteralInt(256)])
    base_int = [
        create_literal_class(
            t_int,
            type_info[t_int],
            refined_type_to_metahandler(RefinedType(xname, t_int, LiquidApp(Name("&&", 0), [gtm1, lt256]))),
        )
    ]

    return [create_literal_class(aeon_ty, type_info[aeon_ty]) for aeon_ty in types] + base_int


def create_literal_ref_nodes(type_info: dict[Type, TypingType] = None) -> list[TypingType]:
    """Creates all literal nodes for refined types, via metahandlers"""
    ref_types = [ty for ty in type_info if isinstance(ty, RefinedType) and ty.type in aeon_to_python]
    return [
        create_literal_class(aeon_ty, type_info[aeon_ty], refined_type_to_metahandler(aeon_ty)) for aeon_ty in ref_types
    ]


def create_var_node(name: Name, ty: Type, python_ty: TypingType) -> TypingType:
    """Creates a python type for a given variable in context."""
    vname = mangle_var(name)
    dc = make_dataclass(f"var_{vname}", [], bases=(python_ty,))

    def get_core(_):
        return Var(name)

    setattr(dc, "get_core", get_core)
    dc = weight(VAR_WEIGHT)(dc)
    return dc


def create_var_apps_node(name: Name, ty: AbstractionType, type_info: dict[Type, TypingType]) -> TypingType:
    """Creates a python type for a given variable in context that is a function."""

    # Collect arguments
    args: list[tuple[Name, Type]] = []
    current: Type = ty
    while isinstance(current, AbstractionType):
        args.append((current.var_name, current.var_type))
        current = current.type
    rtype = current
    python_ty = type_info[rtype]

    vname = mangle_var(name)
    dc = make_dataclass(
        f"var_app_{vname}", [(mangle_name(aname), type_info[ty]) for (aname, ty) in args], bases=(python_ty,)
    )

    def get_core(_self):
        current = Var(name)
        for aname, _ in args:
            current = Application(current, getattr(_self, mangle_name(aname)).get_core())

        return current

    setattr(dc, "get_core", get_core)
    dc = weight(VAR_WEIGHT)(dc)
    return dc


def create_var_nodes(vars: list[Tuple[Name, Type]], type_info: dict[Type, TypingType]) -> list[TypingType]:
    """Creates a list of python types for all variables in context."""
    return [create_var_node(var_name, ty, type_info[ty]) for (var_name, ty) in vars if ty in type_info] + [
        create_var_apps_node(var_name, ty, type_info)
        for (var_name, ty) in vars
        if ty in type_info and isinstance(ty, AbstractionType)
    ]


def create_abstraction_node(ty: AbstractionType, type_info: dict[Type, TypingType]) -> TypingType:
    """Creates a dataclass to represent an abstraction (\\_0 -> x) of type sth_arrow_X."""
    vname = f"lambda_{mangle_type(ty)}"
    dc = make_dataclass(vname, [("body", type_info[ty.type])], bases=(type_info[ty],))

    def get_core(_self):
        return Annotation(Abstraction(Name("_0", fresh_counter.fresh()), _self.body.get_core()), ty)

    # Note: We cannot use the variable inside Abstraction.
    setattr(dc, "get_core", get_core)
    dc = weight(ABS_WEIGHT)(dc)
    return dc


def collect_all_abstractions(t: Type) -> Generator[AbstractionType]:
    match t:
        case RefinedType(_, _, _) | TypeConstructor(_) | TypeVar() | Top():
            return
        case AbstractionType(_, aty, rty):
            yield t
            yield from collect_all_abstractions(aty)
            yield from collect_all_abstractions(rty)
        case TypePolymorphism(_, _, body):
            yield from collect_all_abstractions(body)
        case _:
            assert False, f"Unsupported {t}"


def create_abstraction_nodes(type_info: dict[Type, TypingType]) -> list[TypingType]:
    return [
        create_abstraction_node(ity, type_info)
        for ty in type_info
        for ity in collect_all_abstractions(ty)
        if isinstance(ity, AbstractionType)
    ]


def create_application_node(ty: AbstractionType, type_info: dict[Type, TypingType]) -> TypingType:
    """Creates a dataclass to represent an abstraction (\\_0 -> x) of type sth_arrow_X."""
    vname = f"app_{mangle_type(ty)}"
    dc = make_dataclass(vname, [("fun", type_info[ty]), ("arg", type_info[ty.var_type])], bases=(type_info[ty.type],))

    # Note: this would require dependent type dynamic processing on the return type (parent class)

    def get_core(_self):
        return Application(_self.fun.get_core(), _self.arg.get_core())

    setattr(dc, "get_core", get_core)
    dc = weight(APP_WEIGHT)(dc)
    return dc


def create_application_nodes(type_info: dict[Type, TypingType]) -> list[TypingType]:
    return [create_application_node(ty, type_info) for ty in type_info if isinstance(ty, AbstractionType)]


def create_if_node(ty: Type, type_info: dict[Type, TypingType]) -> TypingType:
    v_name = f"if_{mangle_type(ty)}"
    dc = make_dataclass(
        v_name,
        [("cond", type_info[t_bool]), ("then", type_info[ty]), ("otherwise", type_info[ty])],
        bases=(type_info[ty],),
    )

    def get_core(_self):
        return Annotation(If(_self.cond.get_core(), _self.then.get_core(), _self.otherwise.get_core()), ty)

    setattr(dc, "get_core", get_core)
    dc = weight(IF_WEIGHT)(dc)
    return dc


def create_if_nodes(type_info: dict[Type, TypingType]) -> list[TypingType]:
    return [create_if_node(ty, type_info) for ty in type_info]


def filter_uninterpreted(lt: LiquidTerm) -> Optional[LiquidTerm]:
    match lt:
        case LiquidHole():
            return lt
        case LiquidLiteralBool(_) | LiquidLiteralInt(_) | LiquidLiteralFloat(_) | LiquidLiteralString(_) | LiquidVar(_):
            return lt
        case LiquidApp(fun, args):
            if fun in ["&&", "||"]:
                nargs = [filter_uninterpreted(x) for x in args]
                if nargs[0] is None:
                    return nargs[1]
                if nargs[1] is None:
                    return nargs[0]
                return LiquidApp(fun, nargs)
            elif fun in ops:
                nargs = [filter_uninterpreted(x) for x in args]
                if None not in nargs:
                    return LiquidApp(fun, nargs)
                else:
                    return None
            else:
                # user-defined uninterpreted function
                return None
        case _:
            assert False, f"Failed {lt}"


def remove_uninterpreted_functions_from_type(ty: Type) -> Type:
    match ty:
        case TypeConstructor(_, _) | TypeVar(_) | Top():
            return ty
        case AbstractionType(var_name, var_type, type):
            return AbstractionType(
                var_name,
                remove_uninterpreted_functions_from_type(var_type),
                remove_uninterpreted_functions_from_type(type),
            )
        case RefinedType(name, type, ref):
            ref_filtered = filter_uninterpreted(ref)
            if ref_filtered is None:
                return type
            else:
                return RefinedType(name, type, ref_filtered)
        case TypePolymorphism(name, kind, body):
            return TypePolymorphism(name, kind, remove_uninterpreted_functions_from_type(body))
        case _:
            assert False, f"Unsupported {ty}"


def remove_uninterpreted_functions_wrap(e: TypingContextEntry) -> TypingContextEntry:
    match e:
        case VariableBinder(name, type):
            type = remove_uninterpreted_functions_from_type(type)
            return VariableBinder(name, type)
        case UninterpretedBinder(_, _) | TypeBinder(_, _) | TypeConstructorBinder(_, _):
            return e
        case _:
            assert False, f"Unsupported {e}"


def remove_uninterpreted_functions(ctx: TypingContext) -> TypingContext:
    return TypingContext([remove_uninterpreted_functions_wrap(el) for el in ctx.entries])


def propagate_constants(ctx: TypingContext) -> tuple[TypingContext, dict[Name, LiquidTerm]]:
    substitutions: dict[Name, LiquidTerm] = {}
    entries: list[TypingContextEntry] = []
    for e in ctx.entries:
        match e:
            case VariableBinder(name, ty):
                # Apply all pending substitions

                for k in substitutions:
                    ty = substitution_liquid_in_type(ty, substitutions[k], k)

                # Detect the pattern {n:T | n == C}
                match ty:
                    case RefinedType(name, _, LiquidApp(Name("==", _), [LiquidVar(iname), val])):
                        if name == iname:
                            substitutions[name] = val
                entries.append(VariableBinder(name, ty))
            case _:
                # TODO: we might want to apply the substitions on other types of entries
                entries.append(e)

    return TypingContext(entries), substitutions


def gen_grammar_nodes(
    ctx: TypingContext,
    ty: Type,
    synth_func_name: Name,
    metadata: Metadata,
    grammar_nodes: list[type] | None = None,
) -> tuple[list[type], type]:
    """Generate grammar nodes from the variables in the given TypingContext.

    This function iterates over the variables in the provided TypingContext. For each variable,
    it generates a new class using the create_class_from_ctx_var function and adds it to
    the list of grammar_nodes. If no initial list of grammar_nodes is provided, it starts with an empty list.

    Args:
        ctx (TypingContext): The TypingContext to extract variables from.
        synth_func_name (str) : The name of the function where the hole is located
        metadata (Metadata): The metadata of the program.
        grammar_nodes (list[type]): Initial list of grammar nodes. Defaults to an empty list.

    Returns:
        list[type]: The list of generated grammar nodes.
    """
    if grammar_nodes is None:
        grammar_nodes = []

    # Clean context to remove non-interpreted functions from the context.
    # Simple refinements like 0 < n && n < 10 are kept.
    ctx, _ = propagate_constants(ctx)
    ctx = remove_uninterpreted_functions(ctx)
    ty = remove_uninterpreted_functions_from_type(ty)

    current_metadata = metadata.get(synth_func_name, {})
    is_recursion_allowed = current_metadata.get("recursion", False)
    vars_to_ignore = current_metadata.get("hide", [])
    types_to_ignore = current_metadata.get("hide_types", [])

    def skip(name: Name) -> bool:
        if name == synth_func_name:
            return not is_recursion_allowed
        elif name in vars_to_ignore:
            return True
        elif name.name.startswith("__internal__"):
            return True
        elif name.name in ["native", "native_import", "print"]:
            return True
        else:
            return False

    ctx_vars = [(var_name, ty) for (var_name, ty) in ctx.concrete_vars() if not skip(var_name)]
    types_to_consider = set([t_bool, t_float, t_int, t_string]) | set([x[1] for x in ctx_vars]) | set([ty])
    types_to_consider = types_to_consider - set(TypeConstructor(t) for t in types_to_ignore)
    type_info = extract_all_types(list(types_to_consider))
    type_nodes = list(set(type_info.values()))

    literals = create_literals_nodes(type_info)
    vars = create_var_nodes(ctx_vars, type_info)
    abstractions = create_abstraction_nodes(type_info)
    applications = create_application_nodes(type_info)
    literals_ref = create_literal_ref_nodes(type_info)
    ifs = create_if_nodes(type_info)

    ret = type_nodes + literals + literals_ref + vars + applications + abstractions
    if mangle_name(synth_func_name) in metadata and "disable_control_flow" in metadata[synth_func_name]:
        ret = ret + ifs
    return ret, type_info[ty]


def print_grammar_nodes_names(grammar_nodes: list[type]) -> None:
    class_names = [cls.__name__ for cls in grammar_nodes]
    print(class_names)


def print_grammar_nodes(grammar_nodes: list[type]) -> None:
    for cls in grammar_nodes:
        parents = [base.__name__ for base in cls.__bases__]
        print("@dataclass")
        print(f"class {cls.__name__}({', '.join(parents)}):")
        class_vars = cls.__annotations__
        if class_vars:
            for var_name, var_type in class_vars.items():
                print(f"\t {var_name}: {str(var_type)}")
        else:
            print("\t pass")
        print()
        # print("---------------------------------------------------")


def create_grammar(ctx: TypingContext, ty: Type, fun_name: Name, metadata: Metadata) -> Grammar:
    grammar_nodes, starting_node = gen_grammar_nodes(ctx, ty, fun_name, metadata, [])
    g = extract_grammar(grammar_nodes, starting_node)
    g = g.usable_grammar()
    return g



from aeon.utils.name import Name
from aeon.sugar.stypes import (
    SAbstractionType,
    SRefinedType,
    SType,
    STypeConstructor,
    STypePolymorphism,
    STypeVar,
)
from aeon.sugar.program import (
    SAbstraction,
    SAnnotation,
    SApplication,
    SHole,
    SIf,
    SLet,
    SLiteral,
    SRec,
    STerm,
    STypeAbstraction,
    STypeApplication,
    SVar,
)


def type_equality(a: SType, b: SType, rename_left: dict[Name, Name] | None = None) -> bool:
    "Equality of types up to alpha renaming"
    rename_left = rename_left or {}
    match a, b:
        case STypeVar(an), STypeVar(bn):
            return rename_left.get(an, an) == bn
        case SRefinedType(aname, aty, aref), SRefinedType(bname, bty, bref):
            return type_equality(aty, bty, rename_left | {aname: bname}) and term_equality(
                aref, bref, rename_left | {aname: bname}
            )
        case SAbstractionType(aname, a1, a2), SAbstractionType(bname, b1, b2):
            return type_equality(a1, b1, rename_left) and type_equality(a2, b2, rename_left | {aname: bname})
        case STypeConstructor(aname, a1), STypeConstructor(bname, b1):
            return aname == bname and all([type_equality(a, b, rename_left) for a, b in zip(a1, b1)])
        case STypePolymorphism(aname, akind, abody), STypePolymorphism(bname, bkind, bbody):
            return akind == bkind and type_equality(abody, bbody, rename_left | {aname: bname})
        case STypeApplication(ab, at), STypeApplication(bb, bt):
            return term_equality(ab, bb, rename_left) and type_equality(at, bt, rename_left)
        case STypeAbstraction(aname, akind, abody), STypeAbstraction(bname, bkind, bbody):
            return akind == bkind and term_equality(abody, bbody, rename_left | {aname: bname})
        case _:
            return False


def term_equality(a: STerm, b: STerm, rename_left: dict[Name, Name] | None = None) -> bool:
    "Equality of terms up to alpha renaming"
    rename_left = rename_left or {}
    match a, b:
        case SVar(an), SVar(bn):
            return rename_left.get(an, an) == bn
        case SHole(an), SHole(bn):
            return rename_left.get(an, an) == bn
        case SLiteral(av, at), SLiteral(bv, bt):
            return av == bv and at == bt
        case SApplication(a1, a2), SApplication(b1, b2):
            return term_equality(a1, b1, rename_left) and term_equality(a2, b2, rename_left)
        case SAbstraction(an, abody), SAbstraction(bn, bbody):
            return term_equality(abody, bbody, rename_left | {an: bn})
        case SLet(aname, aval, acont), SLet(bname, bval, bcont):
            return term_equality(aval, bval, rename_left) and term_equality(acont, bcont, rename_left | {aname: bname})
        case SRec(aname, atype, aval, acont), SRec(bname, btype, bval, bcont):
            nrename = rename_left | {aname: bname}
            return (
                term_equality(aval, bval, nrename)
                and type_equality(atype, btype, rename_left)
                and term_equality(acont, bcont, nrename)
            )
        case SAnnotation(ae, at), SAnnotation(be, bt):
            return term_equality(ae, be, rename_left) and type_equality(at, bt, rename_left)
        case SIf(ac, at, ao), SIf(bc, bt, bo):
            return (
                term_equality(ac, bc, rename_left)
                and term_equality(at, bt, rename_left)
                and term_equality(ao, bo, rename_left)
            )
        case STypeApplication(ab, at), STypeApplication(bb, bt):
            return term_equality(ab, bb, rename_left) and type_equality(at, bt, rename_left)
        case STypeAbstraction(aname, akind, abody), STypeAbstraction(bname, bkind, bbody):
            return akind == bkind and term_equality(abody, bbody, rename_left | {aname: bname})
        case _:
            return False



from dataclasses import dataclass, field

from aeon.utils.superscripts import superscript


class FreshCounter:
    counter: int

    def __init__(self):
        self.counter = 0

    def fresh(self) -> int:
        self.counter += 1
        return self.counter


fresh_counter = FreshCounter()


@dataclass(init=False, unsafe_hash=True)
class Name:
    name: str
    id: int = field(default=-1)

    def __init__(self, name: str, id=-1):
        assert isinstance(name, str)
        self.name = str(name).strip()
        self.id = id

    def __str__(self):
        if self.id == 0:
            return self.name
        elif self.id == -1:
            return f"{self.name}?"
        else:
            return f"{self.name}{superscript(str(self.id))}"

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return isinstance(other, Name) and self.name == other.name and self.id == other.id

    def __lt__(self, other):
        return self.id < other.id

    def pretty(self):
        return self.name



from typing import Callable

from aeon.sugar.program import SApplication, SLiteral, STerm, SVar
from aeon.sugar.stypes import STypeConstructor
from aeon.utils.name import Name


def mk_binop(fresh: Callable[[], str], op: Name, a1: STerm, a2: STerm) -> STerm:
    return SApplication(SApplication(SVar(op), a1), a2)


st_top = STypeConstructor(Name("Top", 0))
st_unit = STypeConstructor(Name("Unit", 0))
st_bool = STypeConstructor(Name("Bool", 0))
st_int = STypeConstructor(Name("Int", 0))
st_float = STypeConstructor(Name("Float", 0))
st_string = STypeConstructor(Name("String", 0))

true = SLiteral(True, st_bool)
false = SLiteral(False, st_bool)

i0 = SLiteral(0, st_int)
i1 = SLiteral(1, st_int)
i2 = SLiteral(2, st_int)



from typing import MutableSequence
from aeon.core.liquid import (
    LiquidApp,
    LiquidLiteralBool,
    LiquidLiteralFloat,
    LiquidLiteralInt,
    LiquidLiteralString,
    LiquidTerm,
    LiquidVar,
)
from aeon.core.terms import (
    Abstraction,
    Annotation,
    Application,
    Hole,
    If,
    Let,
    Rec,
    Term,
    TypeAbstraction,
    TypeApplication,
    Var,
    Literal,
)
from aeon.core.types import (
    AbstractionType,
    LiquidHornApplication,
    RefinedType,
    Top,
    Type,
    TypeConstructor,
    TypePolymorphism,
    TypeVar,
)
from aeon.typechecking.context import (
    TypeBinder,
    TypeConstructorBinder,
    TypingContext,
    TypingContextEntry,
    UninterpretedBinder,
    VariableBinder,
)
from aeon.utils.name import Name, fresh_counter


RenamingSubstitions = list[tuple[str, Name]]


def get_last_id(x: str, subs: RenamingSubstitions) -> Name | None:
    for sub, v in subs[::-1]:
        if x == sub:
            return v
    return None


def check_name(name: Name, subs: RenamingSubstitions) -> tuple[Name, RenamingSubstitions]:
    if name.id == -1:
        nname = Name(name.name, fresh_counter.fresh())
        return nname, subs + [(name.name, nname)]
    else:
        return name, subs + [(name.name, name)]


def apply_subs_name(subs: RenamingSubstitions, name: Name) -> Name:
    for sub, v in subs[::-1]:
        if sub == name.name:
            return v
    return name


def bind_lq(liq: LiquidTerm, subs: RenamingSubstitions) -> LiquidTerm:
    match liq:
        case LiquidLiteralBool(_) | LiquidLiteralInt(_) | LiquidLiteralFloat(_) | LiquidLiteralString(_):
            return liq
        case LiquidVar(name, loc):
            return LiquidVar(apply_subs_name(subs, name), loc=loc)
        case LiquidApp(fun, args, loc):
            nfun = apply_subs_name(subs, fun)
            return LiquidApp(nfun, [bind_lq(a, subs) for a in args], loc=loc)
        case LiquidHornApplication(name, argtyps, loc):
            return LiquidHornApplication(apply_subs_name(subs, name), argtyps, loc=loc)
        case _:
            assert False, f"Constructed {liq} ({type(liq)}) not supported."


def bind_ctx(ctx: TypingContext, subs: RenamingSubstitions) -> tuple[TypingContext, RenamingSubstitions]:
    entries: MutableSequence[TypingContextEntry] = []
    subs = []
    for entry in ctx.entries:
        e: TypingContextEntry
        match entry:
            case VariableBinder(name, ty):
                name, subs = check_name(name, subs)
                e = VariableBinder(name, bind_type(ty, subs))
            case UninterpretedBinder(name, ty):
                name, subs = check_name(name, subs)
                nty = bind_type(ty, subs)
                assert isinstance(nty, AbstractionType)
                e = UninterpretedBinder(name, nty)
            case TypeBinder(name, k):
                name, subs = check_name(name, subs)
                e = TypeBinder(name, k)
            case TypeConstructorBinder(name, args):
                name, subs = check_name(name, subs)
                e = TypeConstructorBinder(name, args)
            case _:
                assert False, f"Unique not supported for {ctx} ({type(ctx)})"
        entries.append(e)
    return TypingContext(entries), subs


def bind_type(ty: Type, subs: RenamingSubstitions) -> Type:
    match ty:
        case Top():
            return Top()
        case TypeVar(name, loc):
            return TypeVar(apply_subs_name(subs, name), loc=loc)
        case TypeConstructor(name, args, loc):
            nargs = [bind_type(aty, subs) for aty in args]
            return TypeConstructor(apply_subs_name(subs, name), nargs, loc=loc)

        case AbstractionType(aname, atype, rtype, loc):
            nname, nsubs = check_name(aname, subs)

            return AbstractionType(nname, bind_type(atype, subs), bind_type(rtype, nsubs), loc=loc)
        case RefinedType(name, ty, ref, loc):
            nty = bind_type(ty, subs)
            nname, nsubs = check_name(name, subs)
            nref = bind_lq(ref, nsubs)
            assert isinstance(nty, TypeConstructor) or isinstance(nty, TypeVar) or isinstance(nty, TypeConstructor)
            return RefinedType(nname, nty, nref, loc=loc)
        case TypePolymorphism(name, kind, body, loc):
            name, subs = check_name(name, subs)
            return TypePolymorphism(name, kind, bind_type(body, subs), loc=loc)
        case _:
            assert False, f"Unique not supported for {ty} ({type(ty)})"


def bind_term(t: Term, subs: RenamingSubstitions) -> Term:
    match t:
        case Literal(_, _):
            return t
        case Var(name, loc):
            return Var(apply_subs_name(subs, name), loc=loc)
        case Hole(name, loc):
            name, _ = check_name(name, subs)
            return Hole(name, loc=loc)
        case Annotation(e, ty, loc):
            return Annotation(bind_term(e, subs), bind_type(ty, subs), loc=loc)
        case Application(e1, e2, loc):
            return Application(bind_term(e1, subs), bind_term(e2, subs), loc=loc)
        case Abstraction(name, body, loc):
            name, subs = check_name(name, subs)
            nbody = bind_term(body, subs)
            return Abstraction(name, nbody, loc=loc)
        case TypeApplication(body, ty, loc):
            return TypeApplication(bind_term(body, subs), bind_type(ty, subs), loc=loc)
        case TypeAbstraction(name, kind, body, loc):
            name, subs = check_name(name, subs)
            return TypeAbstraction(name, kind, bind_term(body, subs), loc=loc)
        case If(cond, then, otherwise, loc):
            return If(bind_term(cond, subs), bind_term(then, subs), bind_term(otherwise, subs), loc=loc)
        case Let(name, body, cont, loc):
            name, nsubs = check_name(name, subs)
            return Let(name, bind_term(body, subs), bind_term(cont, nsubs), loc)
        case Rec(name, ty, body, cont, loc):
            name, subs = check_name(name, subs)
            return Rec(name, bind_type(ty, subs), bind_term(body, subs), bind_term(cont, subs), loc)
        case _:
            assert False, f"Unique not supported for {t} ({type(t)})"


def bind_ids(ctx: TypingContext, t: Term) -> tuple[TypingContext, Term]:
    ctx, subs = bind_ctx(ctx, [])
    return ctx, bind_term(t, subs)



from __future__ import annotations

from typing import Callable

from aeon.core.terms import Abstraction
from aeon.core.terms import Annotation
from aeon.core.terms import Application
from aeon.core.terms import If
from aeon.core.terms import Let
from aeon.core.terms import Literal
from aeon.core.terms import Rec
from aeon.core.terms import Term
from aeon.core.terms import Var
from aeon.core.types import t_bool
from aeon.core.types import t_int
from aeon.utils.name import Name


def is_anf(t: Term) -> bool:
    if isinstance(t, Literal):
        return True
    elif isinstance(t, Var):
        return True
    elif isinstance(t, Let):
        return is_anf(t.var_value) and is_anf(t.body)
    elif isinstance(t, Rec):
        return is_anf(t.var_value) and is_anf(t.body)
    elif isinstance(t, Application):
        return is_anf(t.fun) and (isinstance(t.arg, Var) or isinstance(t.arg, Literal))
    elif isinstance(t, If):
        return (
            (
                isinstance(t.cond, Var)
                or isinstance(
                    t.cond,
                    Literal,
                )
            )
            and is_anf(t.then)
            and is_anf(t.otherwise)
        )
    elif isinstance(t, Annotation):
        return is_anf(t.expr)
    elif isinstance(t, Abstraction):
        return is_anf(t.body)
    else:
        assert False


def mk_binop(fresh: Callable[[], str], op: str, a1: Term, a2: Term) -> Term:
    return Application(Application(Var(Name(op, 0)), a1), a2)


true = Literal(True, t_bool)
false = Literal(False, t_bool)

i0 = Literal(0, t_int)
i1 = Literal(1, t_int)
i2 = Literal(2, t_int)



import sklearn.model_selection
from sklearn.ensemble import RandomForestClassifier
from collections import Counter
from imblearn.over_sampling import SMOTE
from sklearn.metrics import confusion_matrix
from sklearn.preprocessing import StandardScaler, MinMaxScaler
import pandas as pd


def train_test_split_pairs(X, y, test_size):
    """
    Splits features X and labels y into training and test sets, then zips them into (x, y) pairs.

    Parameters:
        X (list): List of feature vectors.
        y (list): List of labels.
        test_size (float): Fraction of the dataset to use as the test set.

    Returns:
        tuple: (train, test), where train and test are lists of (x, y) pairs.
    """
    X_train, X_test, y_train, y_test = sklearn.model_selection.train_test_split(X, y, test_size=test_size)
    train = list(zip(X_train, y_train))
    test = list(zip(X_test, y_test))
    return train, test


def balanced(ds):
    """
    Checks if a dataset is balanced (all classes have the same number of samples).

    Parameters:
        ds (tuple): Tuple (X, y), where X is a list of feature vectors and y is a list of labels.

    Returns:
        bool: True if all classes have the same count, False otherwise.
    """
    _, y = ds
    counts = Counter(y)  # count occurrences of each class label
    return len(set(counts.values())) == 1  # ensure all classes have the same count


def load_dataset(filename):
    """
    Loads a dataset from a CSV file, assuming the last column is the label.

    Parameters:
        filename (str): Path to the CSV file.

    Returns:
        tuple: (X, y), where X is a list of feature vectors and y is a list of labels.
    """
    df = pd.read_csv(filename)
    X = df.iloc[:, :-1].values.tolist()
    y = df.iloc[:, -1].tolist()
    return (X, y)


def split_dataset(ds, fraction):
    """
    Splits a dataset into training and test sets according to the given fraction.

    Parameters:
        ds (tuple): Tuple (X, y), where X is a list of feature vectors and y is a list of labels.
        fraction (float): Fraction of the dataset to use as the training set.

    Returns:
        tuple: ((X_train, y_train), (X_test, y_test))
    """
    X, y = ds
    X_train, X_test, y_train, y_test = sklearn.model_selection.train_test_split(X, y, test_size=1 - fraction)
    return ((X_train, y_train), (X_test, y_test))


def train_random_forest(dataset):
    """
    Trains a RandomForestClassifier on the given dataset.

    Parameters:
        dataset (tuple): Tuple (X, y), where X is a list of feature vectors and y is a list of labels.

    Returns:
        RandomForestClassifier: The trained classifier.
    """
    X, y = dataset
    clf = RandomForestClassifier()
    clf.fit(X, y)
    return clf


def smote_oversample(ds):
    """
    Applies SMOTE oversampling to balance the classes in the dataset.

    Parameters:
        ds (tuple): Tuple (X, y), where X is a list of feature vectors and y is a list of labels.

    Returns:
        tuple: (X_res, y_res), the oversampled features and labels as lists.
    """
    X, y = ds
    X_res, y_res = SMOTE().fit_resample(X, y)
    return (X_res.tolist(), y_res.tolist())


def predict(model, ds):
    """
    Uses the given model to predict labels for the features in ds.

    Parameters:
        model: Trained classifier with a predict method.
        ds (tuple): Tuple (X, _), where X is a list of feature vectors.

    Returns:
        list: Predicted labels.
    """
    X, _ = ds
    y_pred = model.predict(X)
    return y_pred.tolist()


def score(model, ds):
    """
    Computes the accuracy score of the model on the given dataset.

    Parameters:
        model: Trained classifier with a score method.
        ds (tuple): Tuple (X, y), where X is a list of feature vectors and y is a list of labels.

    Returns:
        float: Accuracy score.
    """
    X, y = ds
    return float(model.score(X, y))


def confusion_matrix_(model, ds):
    """
    Computes the confusion matrix for the model's predictions on the dataset.

    Parameters:
        model: Trained classifier with a predict method.
        ds (tuple): Tuple (X, y), where X is a list of feature vectors and y is a list of labels.

    Returns:
        list: Confusion matrix as a nested list.
    """
    X, y = ds
    y_pred = model.predict(X)
    cm = confusion_matrix(y, y_pred)
    return cm.tolist()


def standard_scale(ds):
    """
    Scales features in the dataset to have zero mean and unit variance.

    Parameters:
        ds (tuple): Tuple (X, y), where X is a list of feature vectors and y is a list of labels.

    Returns:
        tuple: (X_scaled, y), where X_scaled is the scaled features as a list.
    """
    X, y = ds
    scaler = StandardScaler()
    X_scaled = scaler.fit_transform(X)
    return (X_scaled.tolist(), y)


def minmax_scale(ds):
    """
    Scales features in the dataset to the [0, 1] range.

    Parameters:
        ds (tuple): Tuple (X, y), where X is a list of feature vectors and y is a list of labels.

    Returns:
        tuple: (X_scaled, y), where X_scaled is the scaled features as a list.
    """
    X, y = ds
    scaler = MinMaxScaler()
    X_scaled = scaler.fit_transform(X)
    return (X_scaled.tolist(), y)



from dataclasses import dataclass, field
from itertools import combinations
from aeon.core.types import BaseKind
from aeon.elaboration.context import ElaborationTypingContext
from aeon.elaboration.instantiation import type_substitution
from aeon.facade.api import (
    UnificationFailedError,
    UnificationKindError,
    UnificationSubtypingError,
    UnificationUnknownTypeError,
)
from aeon.sugar.program import (
    SAbstraction,
    SAnnotation,
    SApplication,
    SHole,
    SIf,
    SLet,
    SLiteral,
    SRec,
    STerm,
    STypeAbstraction,
    STypeApplication,
    SVar,
)
from aeon.sugar.stypes import (
    SAbstractionType,
    SRefinedType,
    STypeConstructor,
    STypeVar,
    SType,
    STypePolymorphism,
    get_type_vars,
)
from aeon.sugar.substitutions import substitution_sterm_in_stype
from aeon.utils.name import Name, fresh_counter
from aeon.sugar.ast_helpers import st_top, st_unit, st_bool


def base(ty: SType) -> SType:
    """Returns the base type of a Refined Type"""
    if isinstance(ty, SRefinedType):
        return ty.type
    return ty


@dataclass
class UnificationVar(SType):
    name: Name
    lower: list[SType] = field(default_factory=list)
    upper: list[SType] = field(default_factory=list)

    def constrain_upper(self, ty):
        bty = base(ty)
        self.upper.append(bty)


@dataclass
class Union(SType):
    united: list[SType]


@dataclass
class Intersection(SType):
    intersected: list[SType]


def elaborate_foralls(e: STerm) -> STerm:
    match e:
        case (
            SLiteral(_, _)
            | SHole(_)
            | SVar(_)
            | SAnnotation(_, _)
            | SAbstraction(_, _)
            | SApplication(_, _)
            | SLet(_, _, _)
            | SIf(_, _, _)
            | STypeApplication(_, _)
            | STypeAbstraction(_, _, _)
        ):
            return e
        case SRec(_, _, _, _, loc):
            e1: SRec = e
            if len(get_type_vars(e.var_type)) > 0:
                nt = e.var_type
                nv = e.var_value
                for typevar in get_type_vars(e.var_type):
                    nt = STypePolymorphism(
                        name=typevar.name,
                        kind=BaseKind(),
                        body=nt,
                    )
                    nv = STypeAbstraction(
                        name=typevar.name,
                        kind=BaseKind(),
                        body=nv,
                    )

                e1 = SRec(e.var_name, nt, nv, e.body, loc=loc)
            return e1
        case _:
            assert False, f"Could not elaborate {e} ({type(e)})"


def unify(ctx: ElaborationTypingContext, sub: SType, sup: SType) -> list[SType]:
    match (sub, sup):
        case (_, STypeConstructor(Name("Top", 0))):
            return []
        case (STypeConstructor(subn, subargs), STypeConstructor(supn, supargs)):
            if subn != supn:
                raise UnificationSubtypingError(SHole(Name("todo")), sub, sup)
            elif len(subargs) != len(supargs):
                raise UnificationSubtypingError(SHole(Name("todo")), sub, sup)
            else:
                rt = []
                for subarg, suparg in zip(subargs, supargs):
                    rt += unify(ctx, subarg, suparg)
                return rt
        case (SRefinedType(_, ty, _), _):
            return unify(ctx, ty, sup)
        case (_, SRefinedType(_, ty, _)):
            return unify(ctx, sub, ty)
        case (UnificationVar(_, _, _), _):
            sub.upper.append(sup)
            for l in sub.lower:
                unify(ctx, l, sup)
            return []
        case (_, UnificationVar(_, _, _)):
            sup.lower.append(sub)
            for u in sup.upper:
                unify(ctx, sub, u)
            return []

        case (STypePolymorphism(name, _, body), _):
            u = UnificationVar(ctx.fresh_typevar())
            nty = type_substitution(sub.body, sub.name, u)
            rt = unify(ctx, nty, sup)
            return [nty] + rt
        case (_, STypePolymorphism(name, _, body)):
            u = UnificationVar(ctx.fresh_typevar())
            unify(ctx, sub, type_substitution(body, name, u))
            return []

        case (SAbstractionType(_, sub_vtype, sub_rtype), SAbstractionType(_, sup_vtype, sup_rtype)):
            unify(ctx, sup_vtype, sub_vtype)
            unify(ctx, sub_rtype, sup_rtype)
            return []

        case (STypeVar(subn), STypeVar(supn)):
            if subn == supn:
                return []
            else:
                raise UnificationSubtypingError(SHole(Name("todo")), sub, sup)

        case (STypeConstructor(name1, args), STypeConstructor(name2, args2)):
            if name1 != name2:
                raise UnificationSubtypingError(SHole(Name("todo")), sub, sup)
            elif len(args) != len(args2):
                raise UnificationSubtypingError(
                    SHole(Name("todo")), sub, sup, msg="Type constructor with different number of arguments."
                )
            else:
                for a1, a2 in zip(args, args2):
                    unify(ctx, a1, a2)
                return []

        case _:
            raise UnificationSubtypingError(SHole(Name("todo")), sub, sup)


def unify_single(vname: str, xs: list[SType]) -> SType:
    match [x for x in xs if x != st_top]:
        case []:
            return st_unit
        case other:
            fst = other[0]
            for snd in other[1:]:
                if snd != fst:
                    raise UnificationFailedError(vname, fst, snd)
            return fst


def remove_unions_and_intersections(ctx: ElaborationTypingContext, ty: SType) -> SType:
    match ty:
        case Union(united):
            # TODO: raise better errors
            return unify_single("?", united)
        case Intersection(intersected):
            return unify_single("?", intersected)
        case SAbstractionType(name, vtype, rtype, loc):
            return SAbstractionType(
                var_name=name,
                var_type=remove_unions_and_intersections(ctx, vtype),
                type=remove_unions_and_intersections(ctx, rtype),
                loc=loc,
            )
        case STypePolymorphism(name, kind, body, loc):
            return STypePolymorphism(
                name=name,
                kind=kind,
                body=remove_unions_and_intersections(
                    ctx,
                    body,
                ),
                loc=loc,
            )
        case STypeConstructor(name, args, loc):
            return STypeConstructor(name, [remove_unions_and_intersections(ctx, arg) for arg in args], loc=loc)
        case SRefinedType(name, ity, ref, loc):
            innert = remove_unions_and_intersections(ctx, ity)
            return SRefinedType(name=name, type=innert, refinement=ref, loc=loc)
        case _:
            return ty


def wrap_unification(ctx: ElaborationTypingContext, t: STerm, us: list[SType]) -> STerm:
    nt = t
    for u in us:
        nt = STypeApplication(nt, u)
    return nt


def ensure_not_polymorphism(ctx: ElaborationTypingContext, t: STerm, ty: SType) -> tuple[STerm, SType]:
    while isinstance(ty, STypePolymorphism):
        u = UnificationVar(ctx.fresh_typevar())
        ty = type_substitution(ty.body, ty.name, u)
        t = STypeApplication(t, u)
    return t, ty


def elaborate_synth(ctx: ElaborationTypingContext, t: STerm) -> tuple[STerm, SType]:
    match t:
        case SLiteral(_, ty):
            return (t, ty)
        case SVar(name):
            match ctx.type_of(name):
                case None:
                    raise UnificationUnknownTypeError(t)
                case ty:
                    return (t, ty)
        case SHole(_):
            u = UnificationVar(ctx.fresh_typevar())
            return (t, u)
        case SAnnotation(expr, ty):
            ann = elaborate_check(ctx, expr, ty)
            return (SAnnotation(ann, ty), ty)
        case SAbstraction(name, body):
            u = UnificationVar(ctx.fresh_typevar())
            nctx = ctx.with_var(name, u)
            (t2, bt) = elaborate_synth(nctx, body)
            return (SAbstraction(name, t2), SAbstractionType(name, u, bt))
        case STypeApplication(body, ty):
            (inner, innert) = elaborate_synth(ctx, body)
            assert isinstance(innert, STypePolymorphism)

            u = UnificationVar(ctx.fresh_typevar())
            u.upper.append(ty)
            u.lower.append(ty)

            nty = type_substitution(innert.body, innert.name, u)
            return (STypeApplication(inner, t.type), nty)
        case SLet(name, value, body):
            u = UnificationVar(ctx.fresh_typevar())
            nval = elaborate_check(ctx, value, u)
            (nbody, nbody_type) = elaborate_synth(ctx.with_var(name, u), body)
            return SLet(name, nval, nbody), nbody_type
        case SIf(cond, then, otherwise):
            ncond = elaborate_check(ctx, cond, st_bool)
            nthen, nthen_type = elaborate_synth(ctx, then)
            nelse, nelse_type = elaborate_synth(ctx, otherwise)
            u = UnificationVar(ctx.fresh_typevar())
            unify(ctx, nthen_type, u)
            unify(ctx, nelse_type, u)
            return SIf(ncond, nthen, nelse), u
        case SApplication(fun, arg):
            (nfun, nfun_type) = elaborate_synth(ctx, fun)
            while isinstance(nfun_type, STypePolymorphism):
                u = UnificationVar(ctx.fresh_typevar())
                nfun = STypeApplication(nfun, u)
                nfun_type = type_substitution(nfun_type.body, nfun_type.name, u)

            match nfun_type:
                case SAbstractionType(_, arg_type, return_type, loc):
                    narg = elaborate_check(ctx, arg, arg_type)
                    match narg:
                        case SLiteral(_) | SVar(_):
                            return SApplication(nfun, narg), return_type
                        case _:
                            nname = Name("_anf", fresh_counter.fresh())
                            return SRec(nname, arg_type, narg, SApplication(nfun, SVar(nname)), loc=loc), return_type
                case _:
                    assert False, f"Expected an abstraction type, but got {nfun_type} for {nfun}."
        case _:
            raise UnificationUnknownTypeError(t)


def elaborate_check(ctx: ElaborationTypingContext, t: STerm, ty: SType) -> STerm:
    match (t, ty):
        case (SAbstraction(name, body, loc=loc), SAbstractionType(aname, aty, rty)):
            nctx = ctx.with_var(name, aty)
            nbody = elaborate_check(nctx, body, substitution_sterm_in_stype(rty, SVar(name), aname))
            return SAbstraction(name, nbody, loc=loc)
        case (SLet(name, val, body, loc=loc), _):
            u = UnificationVar(ctx.fresh_typevar())
            nval = elaborate_check(ctx, val, u)
            nctx = ctx.with_var(name, u)
            nbody = elaborate_check(nctx, body, ty)
            return SLet(name, nval, nbody, loc=loc)
        case (SRec(name, vty, val, body, loc=loc), _):
            nctx = ctx.with_var(name, vty)
            nval = elaborate_check(nctx, val, vty)
            nbody = elaborate_check(nctx, body, ty)
            return SRec(name, vty, nval, nbody, loc=loc)
        case (SIf(cond, then, otherwise, loc=loc), _):
            ncond = elaborate_check(ctx, cond, st_bool)
            nthen = elaborate_check(ctx, then, ty)
            nelse = elaborate_check(ctx, otherwise, ty)
            return SIf(ncond, nthen, nelse, loc=loc)
        case (STypeAbstraction(name, kind, body, loc=loc), STypePolymorphism(tname, tkind, tbody)):
            if kind != tkind:
                assert UnificationKindError(t, ty, kind, tkind)
            nctx = ctx.with_typevar(name, kind)
            nty = type_substitution(tbody, tname, STypeVar(name))
            nbody = elaborate_check(nctx, body, nty)
            return STypeAbstraction(name, kind, nbody, loc=loc)
        case (SApplication(fun, arg, loc=loc), _):
            u = UnificationVar(ctx.fresh_typevar())
            nfun = elaborate_check(ctx, fun, SAbstractionType(Name("_", fresh_counter.fresh()), u, ty))
            narg = elaborate_check(ctx, arg, u)
            return SApplication(nfun, narg, loc=loc)
        case _:
            (c, s) = elaborate_synth(ctx, t)
            (c, s) = get_rid_of_polymorphism(ctx, c, s, ty)
            unify(ctx, s, ty)
            return c


def get_rid_of_polymorphism(ctx: ElaborationTypingContext, c: STerm, s: SType, ty: SType) -> tuple[STerm, SType]:
    while isinstance(s, STypePolymorphism) and not isinstance(ty, STypePolymorphism):
        u = UnificationVar(ctx.fresh_typevar())
        c = STypeApplication(c, u)
        s = type_substitution(s.body, s.name, u)
    return (c, s)


def extract_direction(ty: SType) -> set[SType]:
    assert isinstance(ty, SType)
    match ty:
        case UnificationVar(_, lower, upper):
            r: set = set()
            for t in lower + upper:
                r = r.union(extract_direction(t))
            return r
        case _:
            return set([ty])


def replace_unification_variables(
    ctx: ElaborationTypingContext, ty: SType
) -> tuple[SType, list[Union], list[Intersection]]:
    """Removes unification variables, and replaces them with either Union or
    Intersection Type.

    This function returns lists of unions of intersections.
    """
    unions: list[Union] = []
    intersections: list[Intersection] = []

    def go(ctx: ElaborationTypingContext, ty: SType, polarity: bool) -> SType:
        """The recursive part of the function."""
        match ty:
            case STypeVar(_):
                return ty
            case SAbstractionType(name, vty, rty, loc=loc):
                return SAbstractionType(name, go(ctx, vty, not polarity), go(ctx, rty, polarity), loc=loc)
            case SRefinedType(name, ity, ref, loc=loc):
                nt = go(ctx, ity, polarity)
                return SRefinedType(name, nt, ref, loc=loc)
            case STypePolymorphism(name, kind, body, loc=loc):
                return STypePolymorphism(name, kind, go(ctx, body, polarity), loc=loc)
            case STypeConstructor(name, args, loc=loc):
                return STypeConstructor(name, [go(ctx, arg, polarity) for arg in args], loc=loc)

            case UnificationVar(_, _, _):
                return Union(list(extract_direction(ty)))
            case _:
                assert False

    return (go(ctx, ty, True), unions, intersections)


def remove_from_union_and_intersection(
    unions: list[Union],
    intersections: list[Intersection],
    to_be_removed: list[Name],
):
    """Removes all unification vars whose name is in the to_be_removed list."""

    def rem(x: SType) -> bool:
        if isinstance(x, UnificationVar):
            return x.name not in to_be_removed
        else:
            return True

    for i in intersections:
        i.intersected = list(filter(rem, i.intersected))

    for union in unions:
        union.united = list(filter(rem, union.united))


def handle_unification_in_type(ctx: ElaborationTypingContext, ty: SType) -> SType:
    # Source: https://dl.acm.org/doi/pdf/10.1145/3409006
    nt, unions, intersections = replace_unification_variables(ctx, ty)

    # 1. Removal of polar variable
    all_positive = [x.name for u in unions for x in u.united if isinstance(x, UnificationVar)]
    all_negative = [x.name for i in intersections for x in i.intersected if isinstance(x, UnificationVar)]
    to_be_removed = [x for x in all_positive if x not in all_negative] + [
        x for x in all_negative if x not in all_positive
    ]

    # 2. Unification of indistinguishable variables
    for union in unions:
        unifications = [x for x in union.united if isinstance(x, UnificationVar)]
        for a, b in combinations(unifications, 2):
            if all(a in u.united and b in u.united for u in unions):
                to_be_removed.append(max(a.name, b.name))

    for i in intersections:
        unifications = [x for x in i.intersected if isinstance(x, UnificationVar)]
        for a, b in combinations(unifications, 2):
            if all(a in j.intersected and b in j.intersected for j in intersections):
                to_be_removed.append(max(a.name, b.name))

    # 3. Flattening of variable sandwiches
    unifications = [x for union in unions for x in union.united if isinstance(x, UnificationVar)]
    for u in unifications:
        base_types_together_with_u_pos = [
            b for un in unions if u in un.united for b in un.united if not isinstance(b, UnificationVar)
        ]
        base_types_together_with_u_neg = [
            b for i in intersections if u in i.intersected for b in i.intersected if not isinstance(b, UnificationVar)
        ]
        # TODO: I think we need subtyping here.

        if any(bp in base_types_together_with_u_neg for bp in base_types_together_with_u_pos):
            to_be_removed.append(u.name)

    remove_from_union_and_intersection(
        unions,
        intersections,
        to_be_removed,
    )
    return remove_unions_and_intersections(ctx, nt)


def elaborate_remove_unification(ctx: ElaborationTypingContext, t: STerm) -> STerm:
    match t:
        case SLiteral() | SVar() | SHole():
            return t
        case SAnnotation(expr, ty, loc=loc):
            return SAnnotation(elaborate_remove_unification(ctx, expr), ty, loc=loc)
            # TODO NOW: what about ty?
        case SAbstraction(name, body, loc=loc):
            return SAbstraction(name, elaborate_remove_unification(ctx, body), loc=loc)
        case SLet(name, val, body, loc=loc):
            nctx = ctx.with_var(t.var_name, st_unit)  # TODO poly: Unit??
            return SLet(name, elaborate_remove_unification(ctx, val), elaborate_remove_unification(nctx, body), loc=loc)
        case SRec(name, ty, val, body, loc=loc):
            nty = handle_unification_in_type(ctx, ty)
            nt = remove_unions_and_intersections(ctx, ty)
            nctx = ctx.with_var(t.var_name, t.var_type)
            return SRec(
                name, nty, elaborate_remove_unification(nctx, val), elaborate_remove_unification(nctx, body), loc=loc
            )

        case SIf(cond, then, otherwise, loc=loc):
            return SIf(
                elaborate_remove_unification(ctx, cond),
                elaborate_remove_unification(ctx, then),
                elaborate_remove_unification(ctx, otherwise),
                loc=loc,
            )

        case SApplication(fun, arg, loc=loc):
            return SApplication(elaborate_remove_unification(ctx, fun), elaborate_remove_unification(ctx, arg), loc=loc)

        case STypeAbstraction(name, kind, body, loc=loc):
            nctx = ctx.with_typevar(name, kind)
            return STypeAbstraction(name, kind, elaborate_remove_unification(nctx, body), loc=loc)

        case STypeApplication(body, ty, loc=loc):
            # Recursively apply itself.
            body = elaborate_remove_unification(ctx, body)

            nt = handle_unification_in_type(ctx, ty)

            match nt:
                case STypeConstructor(Name("Top", _)):
                    return STypeApplication(body, nt, loc=loc)
                case _:
                    should_be_refined = True
                    match body:
                        case SVar(name):
                            match ctx.type_of(name):
                                case STypePolymorphism(_, BaseKind(), _):
                                    should_be_refined = False
                    match nt:
                        case STypeConstructor(_) | STypeVar(_) | STypeConstructor(_, _):
                            new_type: SType
                            if should_be_refined:
                                nv = Name("self", fresh_counter.fresh())
                                nh = Name("k", fresh_counter.fresh())
                                # TODO NOW: liquidhornapp
                                # ref = LiquidHornApplication("k", [(LiquidVar(new_var), str(nt))])
                                new_type = SRefinedType(nv, nt, SHole(nh))
                            else:
                                new_type = nt
                            return STypeApplication(body, new_type, loc=loc)
                        case SRefinedType(name, ity, ref):
                            # TODO NOW: liquidhornapp
                            # ref = LiquidHornApplication("k", [(LiquidVar(nt.name), str(nt.type))])
                            nh = Name("k", fresh_counter.fresh())
                            ref = SHole(nh)
                            new_type = SRefinedType(name, ity, ref)
                            return STypeApplication(body, new_type, loc=loc)
                        case SAbstractionType(_, _, _):
                            # AbstractionTypes are not refined
                            return STypeApplication(body, nt, loc=loc)
                        # case TypePolymorphism(name, kind, body_):

                        case _:
                            assert False, f"{nt} ({type(nt)}) not an SType."
        case _:
            assert False, f"{t} ({type(t)}) not an STerm."


def elaborate(ctx: ElaborationTypingContext, e: STerm, expected_type: SType = st_top) -> STerm:
    e2 = elaborate_foralls(e)
    e3 = elaborate_check(ctx, e2, expected_type)
    e4 = elaborate_remove_unification(ctx, e3)
    return e4



from __future__ import annotations

import pathlib
from typing import Callable

from lark import Lark, Transformer, v_args

from aeon.core.types import BaseKind, StarKind
from aeon.sugar.program import (
    Decorator,
    SAbstraction,
    SAnnotation,
    SApplication,
    SHole,
    SIf,
    SLet,
    SLiteral,
    SRec,
    STerm,
    STypeAbstraction,
    STypeApplication,
    SVar,
)
from aeon.sugar.program import Definition
from aeon.sugar.program import ImportAe
from aeon.sugar.program import Program
from aeon.sugar.program import TypeDecl
from aeon.sugar.program import InductiveDecl
from aeon.sugar.stypes import (
    SAbstractionType,
    SType,
    STypeConstructor,
    STypeVar,
    SRefinedType,
    STypePolymorphism,
)

from aeon.sugar.ast_helpers import i0
from aeon.sugar.ast_helpers import st_int, st_bool, st_string, st_float, st_unit
from aeon.sugar.stypes import builtin_types
from aeon.utils.name import Name, fresh_counter

from aeon.utils.location import FileLocation, Location


def ensure_list(a):
    if isinstance(a, list):
        return a
    else:
        return [a]


class AnnotatedStr(str):
    loc: Location

    def __init__(self, value: str, loc: Location):
        self.loc = loc


class TreeToSugar(Transformer):
    filename: str
    counter: int

    def __init__(self, filename: str, start_counter: int = 0):
        self.filename = filename
        self.counter = start_counter

    def _loc(self, meta):
        return FileLocation(self.filename, start=(meta.line, meta.column), end=(meta.end_line, meta.end_column))

    def same(self, args):
        return args[0]

    # Types
    def refined_t(self, args):
        return SRefinedType(Name(args[0]), args[1], args[2])

    def abstraction_t(self, args):
        return SAbstractionType(Name(args[0]), args[1], args[2])

    def polymorphism_t(self, args):
        return STypePolymorphism(Name(args[0]), args[1], args[2])

    def simple_t(self, args):
        name_str = str(args[0])
        if name_str in builtin_types:
            name = Name(name_str, 0)
            return STypeConstructor(name)
        else:
            name = Name(name_str)
            return STypeVar(name)

    def constructor_t(self, args):
        return STypeConstructor(Name(args[0]), args[1:])

    # Expressions
    @v_args(meta=True)
    def annotation(self, meta, args):
        return SAnnotation(args[0], args[1], loc=self._loc(meta))

    @v_args(meta=True)
    def minus(self, meta, args):
        # TODO: check for length of args instead?
        if isinstance(args[0], SLiteral) and type(args[0]) is int:
            return SLiteral(-args[0].value, args[0].type, loc=self._loc(meta))
        return self.binop([i0, args[0]], "-", meta)

    @v_args(meta=True)
    def let_e(self, meta, args):
        return SLet(Name(args[0]), args[1], args[2], loc=self._loc(meta))

    @v_args(meta=True)
    def rec_e(self, meta, args):
        return SRec(Name(args[0]), args[1], args[2], args[3], loc=self._loc(meta))

    @v_args(meta=True)
    def if_e(self, meta, args):
        return SIf(args[0], args[1], args[2], loc=self._loc(meta))

    @v_args(meta=True)
    def nnot(self, meta, args):
        return SApplication(SVar(Name("!", 0), loc=self._loc(meta)), args[0], loc=self._loc(meta))

    @v_args(meta=True)
    def binop_eq(self, meta, args):
        return self.binop(args, "==", meta)

    @v_args(meta=True)
    def binop_neq(self, meta, args):
        return self.binop(args, "!=", meta)

    @v_args(meta=True)
    def binop_and(self, meta, args):
        return self.binop(args, "&&", meta)

    @v_args(meta=True)
    def binop_or(self, meta, args):
        return self.binop(args, "||", meta)

    @v_args(meta=True)
    def binop_lt(self, meta, args):
        return self.binop(args, "<", meta)

    @v_args(meta=True)
    def binop_le(self, meta, args):
        return self.binop(args, "<=", meta)

    @v_args(meta=True)
    def binop_gt(self, meta, args):
        return self.binop(args, ">", meta)

    @v_args(meta=True)
    def binop_ge(self, meta, args):
        return self.binop(args, ">=", meta)

    @v_args(meta=True)
    def binop_imp(self, meta, args):
        return self.binop(args, "-->", meta)

    @v_args(meta=True)
    def binop_plus(self, meta, args):
        return self.binop(args, "+", meta)

    @v_args(meta=True)
    def binop_minus(self, meta, args):
        return self.binop(args, "-", meta)

    @v_args(meta=True)
    def binop_mult(self, meta, args):
        return self.binop(args, "*", meta)

    @v_args(meta=True)
    def binop_div(self, meta, args):
        return self.binop(args, "/", meta)

    @v_args(meta=True)
    def binop_mod(self, meta, args):
        return self.binop(args, "%", meta)

    def binop(self, args, op: str, meta):
        return SApplication(
            SApplication(SVar(Name(op, 0), loc=self._loc(meta)), args[0], loc=self._loc(meta)),
            args[1],
            loc=self._loc(meta),
        )

    @v_args(meta=True)
    def application_e(self, meta, args):
        return SApplication(args[0], args[1], loc=self._loc(meta))

    @v_args(meta=True)
    def abstraction_e(self, meta, args):
        return SAbstraction(Name(args[0]), args[1], loc=self._loc(meta))

    @v_args(meta=True)
    def tabstraction_e(self, meta, args):
        return STypeAbstraction(Name(args[0]), args[1], args[2], loc=self._loc(meta))

    @v_args(meta=True)
    def type_application_e(self, meta, args):
        return STypeApplication(args[0], args[1], loc=self._loc(meta))

    @v_args(meta=True)
    def var(self, meta, args):
        return SVar(Name(args[0]), loc=self._loc(meta))

    @v_args(meta=True)
    def hole(self, meta, args):
        return SHole(Name(args[0]), loc=self._loc(meta))

    @v_args(meta=True)
    def int_lit(self, meta, args):
        return SLiteral(int(args[0]), type=st_int, loc=self._loc(meta))

    @v_args(meta=True)
    def float_lit(self, meta, args):
        return SLiteral(float(args[0]), type=st_float, loc=self._loc(meta))

    @v_args(meta=True)
    def bool_lit(self, meta, args):
        value = str(args[0]) == "true"
        return SLiteral(value, type=st_bool, loc=self._loc(meta))

    @v_args(meta=True)
    def string_lit(self, meta, args):
        return SLiteral(args[0], type=st_string, loc=self._loc(meta))

    def ESCAPED_STRING(self, val):
        # TODO: This is terrible and doesn't handle escapes
        v = str(val)[1:-1]
        return v

    def base_kind(self, args):
        return BaseKind()

    def star_kind(self, args):
        return StarKind()

    def list(self, args):
        return args

    def program(self, args):
        non_inductive = [el for el in args[1] if not isinstance(el, InductiveDecl)]
        inductive = [el for el in args[1] if isinstance(el, InductiveDecl)]
        return Program(args[0], non_inductive, inductive, args[2])

    @v_args(meta=True)
    def regular_imp(self, meta, args):
        return ImportAe(args[0], [], loc=self._loc(meta))

    @v_args(meta=True)
    def function_imp(self, meta, args):
        return ImportAe(args[1], args[0], loc=self._loc(meta))

    @v_args(meta=True)
    def type_decl(self, meta, args):
        return TypeDecl(Name(args[0]), [], loc=self._loc(meta))

    @v_args(meta=True)
    def type_constructor_decl(self, meta, args):
        return TypeDecl(Name(args[0]), [Name(i) for i in args[1:]], loc=self._loc(meta))

    @v_args(meta=True)
    def inductive(self, meta, args):
        return InductiveDecl(
            Name(args[0]), [Name(i) for i in args[1]], ensure_list(args[2]), ensure_list(args[3]), loc=self._loc(meta)
        )

    @v_args(meta=True)
    def def_ind_cons(self, meta, args):
        return Definition(Name(args[0]), [], args[1], args[2], SLiteral(None, st_unit), loc=self._loc(meta))

    @v_args(meta=True)
    def def_cons(self, meta, args):
        if len(args) == 3:
            return Definition(Name(args[0]), [], [], args[1], args[2], loc=self._loc(meta))
        else:
            decorators = args[0]
            return Definition(Name(args[1]), [], [], args[2], args[3], decorators, loc=self._loc(meta))

    @v_args(meta=True)
    def def_fun(self, meta, args):
        if len(args) == 4:
            return Definition(Name(args[0]), [], args[1], args[2], args[3], loc=self._loc(meta))
        else:
            decorators = args[0]
            return Definition(Name(args[1]), [], args[2], args[3], args[4], decorators, loc=self._loc(meta))

    def macros(self, args):
        return args

    @v_args(meta=True)
    def macro(self, meta, args):
        return Decorator(Name(args[0]), args[1], loc=self._loc(meta))

    def macro_args(self, args):
        return args

    def empty_list(self, args):
        return []

    def args(self, args):
        return args

    def arg(self, args):
        return (Name(args[0]), args[1])

    def refined_arg(self, args):
        return Name(args[0]), SRefinedType(Name(args[0]), args[1], args[2])

    def abstraction_refined_t(self, args):
        type = SRefinedType(Name(args[0]), args[1], args[2])
        return SAbstractionType(Name(args[0]), type, args[3])

    def abstraction_et(self, args):
        return SAnnotation(
            SAbstraction(Name(args[0]), args[2]),
            SAbstractionType(
                Name(args[0]), args[1], STypeVar(Name("?t", fresh_counter.fresh()))
            ),  # TODO NOW: understand this?
        )


def mk_parser(rule="start", start_counter: int = 0, filename: str = ""):
    parser = Lark.open(
        pathlib.Path(__file__).parent.absolute() / "aeon_sugar.lark",
        parser="lalr",
        # lexer='standard',
        start=rule,
        import_paths=[pathlib.Path(__file__).parent.parent.absolute() / "frontend"],
        propagate_positions=True,
    )
    transf = TreeToSugar(filename, start_counter)

    def parse(s: str):
        tree = parser.parse(s)
        return transf.transform(tree)

    return parse


parse_expression: Callable[[str], STerm] = mk_parser("expression")
parse_program: Callable[[str], Program] = mk_parser("program")
parse_type: Callable[[str], SType] = mk_parser("type", filename="<prelude>")


def parse_main_program(source: str, filename: str) -> Program:
    return mk_parser("program", filename=filename)(source)



from aeon.core.liquid import LiquidVar
from aeon.core.substitutions import substitution_in_liquid
from aeon.core.types import AbstractionType, TypeConstructor, Type, TypeVar, t_bool, t_unit
from aeon.typechecking.context import TypingContext
from aeon.typechecking.liquid import LiquidTypeCheckingContext, check_liquid, lower_abstraction_type, lower_context
from aeon.verification.smt import rename_constraint
from aeon.verification.vcs import (
    Constraint,
    Implication,
    LiquidConstraint,
    Top,
    Conjunction,
    TypeVarDeclaration,
    UninterpretedFunctionDeclaration,
)
from aeon.utils.name import Name, fresh_counter


def wellformed_constraint_liquid(ctx: LiquidTypeCheckingContext, c: Constraint) -> None:
    match c:
        case LiquidConstraint(expr):
            assert check_liquid(ctx, expr, t_bool), f"Expected boolean expression, got {expr}."
        case Conjunction(c1, c2):
            wellformed_constraint_liquid(ctx, c1)
            wellformed_constraint_liquid(ctx, c2)
        case UninterpretedFunctionDeclaration(name, ft, seq):
            nctx = LiquidTypeCheckingContext(
                ctx.known_types, ctx.variables, ctx.functions | {name: lower_abstraction_type(ft)}
            )
            assert isinstance(ft, AbstractionType), f"{ft} is expected to be an AbstractionType."
            wellformed_constraint_liquid(nctx, seq)
        case TypeVarDeclaration(name, seq):
            nctx = LiquidTypeCheckingContext(ctx.known_types + [TypeConstructor(name)], ctx.variables, ctx.functions)
            assert name in ctx.known_types, f"Type {name} not found in known types ({ctx.known_types})."
            wellformed_constraint_liquid(nctx, seq)
        case Implication(name, base, pred, seq):
            if isinstance(base, Top):
                assert False, "Top type is not allowed in Implication."
            nctx = LiquidTypeCheckingContext(ctx.known_types, ctx.variables | {name: base}, ctx.functions)
            assert isinstance(base, TypeConstructor), f"Expected TypeConstructor, got {base}."
            assert name not in ctx.variables, f"Variable {name} already exists in context."

            assert check_liquid(nctx, pred, t_bool), f"Predicate {pred} is either not valid, or not of type boolean."
            wellformed_constraint_liquid(nctx, seq)
        case _:
            assert False, f"Wellformeness of constraint {c} ({type(c)}) not defined"


def canonicalize_type(t: Type) -> Type:
    match t:
        case Top():
            return t_unit
        case TypeConstructor(_, _):
            return t
        case TypeVar(tv):
            return TypeConstructor(tv)
        case AbstractionType(name, at, rt):
            return AbstractionType(name, canonicalize_type(at), canonicalize_type(rt))
        case _:
            assert False, f"No canonicalization for type {t} ({type(t)})"


def canonicalize_constraint(c: Constraint, stack: list[Name] = None) -> Constraint:
    if stack is None:
        stack = []

    match c:
        case LiquidConstraint(expr):
            return LiquidConstraint(expr)
        case Conjunction(c1, c2):
            return Conjunction(canonicalize_constraint(c1, stack), canonicalize_constraint(c2, stack))
        case UninterpretedFunctionDeclaration(name, ft, seq):
            new_name = Name(name.name, fresh_counter.fresh())
            if new_name != name:
                nseq = rename_constraint(c, name, new_name)
            else:
                nseq = seq
            nseq = canonicalize_constraint(nseq, stack + [new_name])
            nty = canonicalize_type(ft)
            assert isinstance(nty, AbstractionType), f"Expected AbstractionType, got {nty}."
            return UninterpretedFunctionDeclaration(new_name, nty, nseq)
        case TypeVarDeclaration(name, seq):
            if name in stack:
                assert False, f"Type variable {name} is already in the stack."
            return TypeVarDeclaration(name, canonicalize_constraint(seq, stack + [name]))
        case Implication(name, base, pred, seq):
            new_name = Name(name.name, fresh_counter.fresh())
            if new_name != name:
                npred = substitution_in_liquid(pred, LiquidVar(new_name), name)
                nseq = rename_constraint(c, name, new_name)
            else:
                npred = pred
                nseq = seq
            nseq = canonicalize_constraint(nseq, stack + [new_name])
            nty = canonicalize_type(base)
            assert isinstance(nty, TypeConstructor) or isinstance(nty, Top) or isinstance(nty, TypeVar), (
                f"Expected TypeConstructor, got {nty}."
            )

            return Implication(new_name, nty, npred, nseq)
        case _:
            assert False, f"Canonicalization of constraint {c} ({type(c)}) not defined"


def wellformed_constraint(ctx: TypingContext, c: Constraint) -> bool:
    lctx = lower_context(ctx)
    wellformed_constraint_liquid(lctx, c)
    return True



from aeon.core.substitutions import substitute_vartype_in_term, substitution
from aeon.core.terms import (
    Abstraction,
    Annotation,
    Application,
    Hole,
    If,
    Let,
    Literal,
    Rec,
    Term,
    TypeAbstraction,
    TypeApplication,
    Var,
)
from aeon.core.types import t_bool, t_int
from aeon.utils.name import Name


def nf(t: Term) -> Term:
    match t:
        case Application(Abstraction(var_name, body), arg):
            return substitution(body, arg, var_name)

        case Application(Annotation(Abstraction(var_name, body), ty), arg):
            return substitution(body, arg, var_name)

        case TypeApplication(TypeAbstraction(vty, kind, body), ty):
            return substitute_vartype_in_term(body, ty, vty)

        case If(Literal(True, _), then, _):
            return then

        case If(Literal(False, _), _, otherwise):
            return otherwise

        # Basic opts

        case Application(Application(Var(Name("&&", _)), Literal(True, tb)), e):
            return e
        case Application(Application(Var(Name("&&", _)), Literal(False, tb)), e):
            return Literal(False, tb)
        case Application(Application(Var(Name("&&", _)), e), Literal(True, tb)):
            return e
        case Application(Application(Var(Name("&&", _)), e), Literal(False, tb)):
            return Literal(False, tb)

        case Application(Application(Var(Name("||", _)), Literal(True, tb)), e):
            return Literal(True, tb)
        case Application(Application(Var(Name("||", _)), Literal(False, tb)), e):
            return e
        case Application(Application(Var(Name("||", _)), e), Literal(True, tb)):
            return Literal(True, tb)
        case Application(Application(Var(Name("||", _)), e), Literal(False, tb)):
            return e

        case Application(Application(Var(Name("+", _)), Literal(0, ti)), e):
            return e
        case Application(Application(Var(Name("+", _)), e), Literal(0, ti)):
            return e
        case Application(Application(Var(Name("+", _)), Literal(a, ti)), Literal(b, tb)):
            return Literal(a + b, ti)  # type: ignore

        case Application(Application(Var(Name("-", _)), e), Literal(0, ti)):
            return e
        case Application(Application(Var(Name("-", _)), Literal(a, ti)), Literal(b, tb)):
            return Literal(a - b, ti)  # type: ignore
        case Application(Application(Var(Name("-", _)), x1), x2):
            if x1 == x2:
                return Literal(0, t_int)
            else:
                return t

        case Application(Application(Var(Name("*", _)), Literal(0, ti)), e):
            return Literal(0, ti)
        case Application(Application(Var(Name("*", _)), e), Literal(0, ti)):
            return Literal(0, ti)
        case Application(Application(Var(Name("*", _)), Literal(1, ti)), e):
            return e
        case Application(Application(Var(Name("*", _)), e), Literal(1, ti)):
            return e
        case Application(Application(Var(Name("*", _)), Literal(a, ti)), Literal(b, tb)):
            return Literal(a * b, ti)  # type: ignore

        case Application(Application(Var(Name("/", _)), Literal(0, ti)), e):
            return Literal(0, ti)

        case Application(Application(Var(Name("/", _)), x1), x2):
            if x1 == x2:
                return Literal(1, t_int)
            else:
                return t

        case Application(Application(Var(Name("%", _)), Literal(0, ti)), _):
            return Literal(0, t_int)

        case Application(Application(Var(Name("%", _)), x1), x2):
            if x1 == x2:
                return Literal(0, t_int)
            else:
                return t

        case Application(Application(Var(Name("==", _)), Literal(a, ti)), Literal(b, tb)):
            return Literal(a == b, t_bool)

        case Application(Application(Var(Name("!=", _)), Literal(a, ti)), Literal(b, tb)):
            return Literal(a != b, t_bool)

        case Application(Application(Var(Name(">", _)), Literal(a, ti)), Literal(b, tb)):
            return Literal(a > b, t_bool)  # type: ignore

        case Application(Application(Var(Name(">=", _)), Literal(a, ti)), Literal(b, tb)):
            return Literal(a >= b, t_bool)  # type: ignore

        case Application(Application(Var(Name("<", _)), Literal(a, ti)), Literal(b, tb)):
            return Literal(a < b, t_bool)  # type: ignore

        case Application(Application(Var(Name("<=", _)), Literal(a, ti)), Literal(b, tb)):
            return Literal(a <= b, t_bool)  # type: ignore

        case Literal(_, _):
            return t
        case Var(_):
            return t
        case Annotation(_, _):
            return t
        case Hole(_):
            return t

        case Abstraction(var_name, body):
            return Abstraction(var_name, nf(body))
        case Application(fun, arg):
            return Application(nf(fun), nf(arg))
        case Let(var_name, var_value, body):
            return substitution(body, nf(var_value), var_name)
        case Rec(var_name, var_type, var_value, body):
            return Rec(var_name, var_type, nf(var_value), nf(body))

        case If(cond, then, otherwise):
            return If(nf(cond), nf(then), nf(otherwise))

        case TypeAbstraction(ty, kind, body):
            return TypeAbstraction(ty, kind, nf(body))
        case TypeApplication(body, ty):
            return TypeApplication(nf(body), ty)
        case _:
            assert False, f"No case for {t} ({type(t)})"


def normal_form(t: Term) -> Term:
    while True:
        nt = nf(t)
        if t == nt:
            return nt
        t = nt


def optimize(t: Term) -> Term:
    return normal_form(t)



from __future__ import annotations
from typing import Iterable

from loguru import logger

from aeon.core.instantiation import type_substitution
from aeon.core.liquid import LiquidApp, LiquidTerm
from aeon.core.types import LiquidHornApplication, StarKind
from aeon.core.liquid import LiquidLiteralBool
from aeon.core.liquid import LiquidLiteralFloat
from aeon.core.liquid import LiquidLiteralInt
from aeon.core.liquid import LiquidVar
from aeon.core.liquid_ops import ops
from aeon.core.substitutions import liquefy, substitute_vartype
from aeon.core.substitutions import substitution_in_type
from aeon.core.terms import Abstraction
from aeon.core.terms import Annotation
from aeon.core.terms import Application
from aeon.core.terms import Hole
from aeon.core.terms import If
from aeon.core.terms import Let
from aeon.core.terms import Literal
from aeon.core.terms import Rec
from aeon.core.terms import Term
from aeon.core.terms import TypeAbstraction
from aeon.core.terms import TypeApplication
from aeon.core.terms import Var
from aeon.core.types import AbstractionType, Kind, is_bare
from aeon.core.types import BaseKind
from aeon.core.types import TypeConstructor
from aeon.core.types import RefinedType
from aeon.core.types import Type
from aeon.core.types import TypePolymorphism
from aeon.core.types import TypeVar
from aeon.core.types import t_bool
from aeon.core.types import t_float
from aeon.core.types import t_int
from aeon.core.types import top
from aeon.core.types import type_free_term_vars
from aeon.facade.api import (
    AeonError,
    CoreInvalidApplicationError,
    CoreSubtypingError,
    CoreTypeApplicationRequiresBareTypesError,
    CoreTypeCheckingError,
    CoreTypingRelation,
    CoreVariableNotInContext,
    CoreWellformnessError,
    CoreWrongKindInTypeApplicationError,
    LiquidTypeCheckingFailedRelation,
)
from aeon.typechecking.context import TypingContext
from aeon.typechecking.entailment import entailment, entailment_context
from aeon.typechecking.well_formed import wellformed
from aeon.verification.horn import fresh
from aeon.verification.sub import ensure_refined
from aeon.verification.sub import implication_constraint
from aeon.verification.sub import sub
from aeon.verification.vcs import Conjunction
from aeon.verification.vcs import Constraint
from aeon.verification.vcs import LiquidConstraint
from aeon.verification.horn import solve
from aeon.utils.name import Name, fresh_counter
from aeon.verification.helpers import (
    remove_unrelated_context,
    simplify_constraint,
    conjunctive_normal_form,
    is_implication_true,
)

ctrue = LiquidConstraint(LiquidLiteralBool(True))


def is_compatible(a: Kind, b: Kind):
    """Returns whether kind a is a subkind of kind b"""
    return (a == b) or b == StarKind()


def argument_is_typevar(ty: Type):
    return (
        isinstance(ty, TypeVar)
        or isinstance(
            ty,
            RefinedType,
        )
        and isinstance(ty.type, TypeVar)
    )


def prim_litbool(t: bool) -> RefinedType:
    vname = Name("v", fresh_counter.fresh())
    if t:
        return RefinedType(vname, t_bool, LiquidVar(vname))
    else:
        return RefinedType(vname, t_bool, LiquidApp(Name("!", 0), [LiquidVar(vname)]))


def prim_litint(t: int) -> RefinedType:
    vname = Name("v", fresh_counter.fresh())
    return RefinedType(
        vname,
        t_int,
        LiquidApp(Name("==", 0), [LiquidVar(vname), LiquidLiteralInt(t)]),
    )


def prim_litfloat(t: float) -> RefinedType:
    vname = Name("v", fresh_counter.fresh())
    return RefinedType(
        vname,
        t_float,
        LiquidApp(Name("==", 0), [LiquidVar(vname), LiquidLiteralFloat(t)]),
    )


def make_binary_app_type(t: Name, ity: TypeConstructor | TypeVar, oty: TypeConstructor | TypeVar) -> Type:
    """Creates the type of a binary operator"""
    xname = Name("x", fresh_counter.fresh())
    yname = Name("y", fresh_counter.fresh())
    zname = Name("z", fresh_counter.fresh())
    output = RefinedType(
        zname,
        oty,
        LiquidApp(
            Name("==", 0),
            [
                LiquidVar(zname),
                LiquidApp(
                    t,
                    [LiquidVar(xname), LiquidVar(yname)],
                ),
            ],
        ),
    )
    appt2 = AbstractionType(yname, ity, output)
    appt1 = AbstractionType(xname, ity, appt2)
    return appt1


def prim_op(t: Name) -> Type:
    match t.name:
        case "%":
            return make_binary_app_type(t, t_int, t_int)
        case "+" | "-" | "*" | "/":
            name_a = Name("a", fresh_counter.fresh())
            return TypePolymorphism(name_a, BaseKind(), make_binary_app_type(t, TypeVar(name_a), TypeVar(name_a)))
        case "==" | "!=" | ">" | ">=" | "<" | "<=":
            name_a = Name("a", fresh_counter.fresh())
            return TypePolymorphism(name_a, BaseKind(), make_binary_app_type(t, TypeVar(name_a), t_bool))
        case "&&" | "||":
            return make_binary_app_type(t, t_bool, t_bool)
        case "!":
            name = Name("fresh", fresh_counter.fresh())
            return AbstractionType(name, t_bool, t_bool)
        case _:
            assert False, f"Unknown selfication of {t}"


def rename_liquid_term(refinement: LiquidTerm, old_name: Name, new_name: Name):
    if isinstance(refinement, LiquidVar):
        if refinement.name == old_name:
            return LiquidVar(new_name)
        else:
            return refinement
    elif isinstance(refinement, LiquidLiteralBool):
        return refinement
    elif isinstance(refinement, LiquidLiteralInt):
        return refinement
    elif isinstance(refinement, LiquidLiteralFloat):
        return refinement
    elif isinstance(refinement, LiquidApp):
        return LiquidApp(
            refinement.fun,
            [rename_liquid_term(x, old_name, new_name) for x in refinement.args],
        )
    elif isinstance(refinement, LiquidHornApplication):
        if refinement.name == old_name:
            return LiquidHornApplication(
                new_name,
                [(rename_liquid_term(x, old_name, new_name), t) for (x, t) in refinement.argtypes],
            )
        else:
            return LiquidHornApplication(
                refinement.name,
                [(rename_liquid_term(x, old_name, new_name), t) for (x, t) in refinement.argtypes],
            )
    else:
        assert False


def renamed_refined_type(ty: RefinedType) -> RefinedType:
    old_name = ty.name
    new_name = Name("_inner_" + old_name.name, fresh_counter.fresh())

    refinement = rename_liquid_term(ty.refinement, old_name, new_name)
    return RefinedType(new_name, ty.type, refinement)


# patterm matching term
def synth(ctx: TypingContext, t: Term) -> tuple[Constraint, Type]:
    match t:
        case Literal(_, TypeConstructor(Name("Unit", _))):
            # TODO: Unit is encoded as True, replace with custom Sort
            return (ctrue, prim_litbool(True))
        case Literal(vb, TypeConstructor(Name("Bool", _))):
            assert isinstance(vb, bool)
            return (ctrue, prim_litbool(vb))
        case Literal(vi, TypeConstructor(Name("Int", _))):
            assert isinstance(vi, int)
            return (ctrue, prim_litint(vi))
        case Literal(vf, TypeConstructor(Name("Float", _))):
            assert isinstance(vf, float)
            return (ctrue, prim_litfloat(vf))
        case Literal(_, TypeConstructor(Name("String", _))):
            # TODO: String support
            return (ctrue, t.type)
        case Var(name):
            if name in ops:
                return (ctrue, prim_op(name))
            ty = ctx.type_of(name)
            if not ty:
                raise CoreVariableNotInContext(ctx, t)
            if isinstance(ty, TypeConstructor) or isinstance(ty, RefinedType) or isinstance(ty, TypeVar):
                ty = ensure_refined(ty)
                assert isinstance(ty, RefinedType)
                # assert ty.name != t.name
                if ty.name == t.name:
                    ty = renamed_refined_type(ty)
                # Self
                ty = RefinedType(
                    ty.name,
                    ty.type,
                    LiquidApp(
                        Name("&&", 0),
                        [
                            ty.refinement,
                            LiquidApp(
                                Name("==", 0),
                                [
                                    LiquidVar(ty.name),
                                    LiquidVar(t.name),
                                ],
                            ),
                        ],
                    ),
                )
            return (ctrue, ty)
        case Application(fun, arg):
            (c, ty) = synth(ctx, fun)
            match ty:
                case AbstractionType(aname, atype, rtype):
                    cp = check(ctx, arg, atype)
                    t_subs = substitution_in_type(rtype, arg, aname)
                    c0 = Conjunction(c, cp)
                    # if ctx.has_uninterpreted_fun(aname):
                    #     selfification = LiquidConstraint(LiquidApp(
                    #             Name("==", 0),
                    #             [
                    #                 LiquidVar(aname),
                    #                 LiquidApp(fun, []),
                    #             ],
                    #     ))
                    #     c0 = Conjunction(c0, selfification)
                    return (c0, t_subs)
                case _:
                    raise CoreInvalidApplicationError(t, ty)
        case Let(var_name, var_value, body):
            (c1, t1) = synth(ctx, var_value)
            nctx: TypingContext = ctx.with_var(var_name, t1)
            (c2, t2) = synth(nctx, body)
            term_vars = type_free_term_vars(t1)
            assert t.var_name not in term_vars
            r = (Conjunction(c1, implication_constraint(var_name, t1, c2)), t2)
            return r
        case Rec(var_name, var_type, var_value, body):
            nrctx: TypingContext = ctx.with_var(var_name, var_type)
            c1 = check(nrctx, var_value, var_type)
            (c2, t2) = synth(nrctx, body)
            c1 = implication_constraint(var_name, var_type, c1)
            c2 = implication_constraint(var_name, var_type, c2)
            return Conjunction(c1, c2), t2
        case Annotation(expr, ty):
            nty = fresh(ctx, ty)
            c = check(ctx, expr, nty)
            return c, nty
        case TypeApplication(body, ty):
            if not is_bare(ty):
                # Type Application only works on bare types.
                raise CoreTypeApplicationRequiresBareTypesError(t, ty)
            (c, tabs) = synth(ctx, body)
            nty = fresh(ctx, ty)
            k = ctx.kind_of(nty)
            if isinstance(nty, RefinedType) and isinstance(nty.refinement, LiquidHornApplication):
                nty = nty.type
                k = ctx.kind_of(nty)
            if isinstance(tabs, TypePolymorphism):
                s = type_substitution(tabs.body, tabs.name, nty)
                if k is None or not is_compatible(k, tabs.kind):
                    raise CoreWrongKindInTypeApplicationError(term=t, type=nty, expected_kind=tabs.kind, actual_kind=k)
                return (c, s)
            else:
                assert isinstance(tabs, AbstractionType)
                return (c, tabs)
        case Hole(name):
            name_a = Name(name.name, fresh_counter.fresh())
            return ctrue, TypePolymorphism(name_a, StarKind(), TypeVar(name_a))  # TODO poly: check kind
        case _:
            logger.log("SYNTH_TYPE", ("Unhandled:", t))
            logger.log("SYNTH_TYPE", ("Unhandled:", type(t)))
            assert False, f"Unhandled term {t} in synth. Type: {type(t)}"


def check(ctx: TypingContext, t: Term, ty: Type) -> Constraint:
    try:
        assert wellformed(ctx, ty)
    except AssertionError:
        raise CoreWellformnessError(ty)
    match t, ty:
        case Abstraction(name, body), AbstractionType(var_name, var_type, ret):
            ret = substitution_in_type(ret, Var(name), var_name)
            c = check(ctx.with_var(name, var_type), body, ret)
            return implication_constraint(name, var_type, c)
        case Let(name, val, body), _:
            (c1, t1) = synth(ctx, val)
            nctx: TypingContext = ctx.with_var(name, t1)
            c2 = check(nctx, body, ty)
            return Conjunction(c1, implication_constraint(name, t1, c2))
        case Rec(var_name, var_type, var_value, body), _:
            t1 = fresh(ctx, var_type)
            nrctx: TypingContext = ctx.with_var(var_name, t1)
            c1 = check(nrctx, var_value, var_type)
            c2 = check(nrctx, body, ty)
            c1 = implication_constraint(var_name, t1, c1)
            c2 = implication_constraint(var_name, t1, c2)
            return Conjunction(c1, c2)
        case If(cond, then, otherwise), _:
            y = Name("_cond", fresh_counter.fresh())
            liq_cond = liquefy(cond)
            assert liq_cond is not None
            if not check_type(ctx, cond, t_bool):
                raise CoreTypingRelation(ctx, cond, t_bool)
            c0 = check(ctx, cond, t_bool)
            name_pos = Name("branch_pos", fresh_counter.fresh())
            c1 = implication_constraint(
                y,
                RefinedType(name_pos, t_int, liq_cond),
                check(ctx, then, ty),
            )
            name_neg = Name("branch_neg", fresh_counter.fresh())
            c2 = implication_constraint(
                y,
                RefinedType(name_neg, t_int, LiquidApp(Name("!", 0), [liq_cond])),
                check(ctx, otherwise, ty),
            )
            return Conjunction(c0, Conjunction(c1, c2))
        case TypeAbstraction(name, kind, body), TypePolymorphism(var_name, var_kind, var_body):
            if var_kind == BaseKind() and kind != var_kind:
                raise CoreWrongKindInTypeApplicationError(
                    term=t,
                    type=ty,
                    actual_kind=var_kind,
                    expected_kind=var_kind,
                )
            itype = substitute_vartype(var_body, TypeVar(name), var_name)
            return check(ctx.with_typevar(name, var_kind), body, itype)
        case _:
            (c, s) = synth(ctx, t)
            cp = sub(ctx, s, ty)
            if cp == LiquidConstraint(LiquidLiteralBool(False)):
                raise CoreSubtypingError(ctx, t, s, ty)
            return Conjunction(c, cp)


def check_type(ctx: TypingContext, t: Term, ty: Type = top) -> bool:
    """Returns whether expression t has type ty in context ctx."""
    try:
        assert wellformed(ctx, ty)
        constraint = check(ctx, t, ty)
        # TODO: convert constraint to canonical form
        # constraint = canonicalize_constraint(constraint, [name for (name, _) in ctx.vars()])
        # assert wellformed_constraint(ctx, constraint), f"Constraint {constraint} not wellformed."
        v = entailment(ctx, constraint)
        return v
    except CoreTypeCheckingError:
        return False


def constraint_to_parts(c: Constraint) -> Iterable[Constraint]:
    """Prepares a constraint into a list of sub-problems for error messages"""
    for cons in conjunctive_normal_form(c):
        if not is_implication_true(cons):
            if not solve(cons):
                cons_simp = simplify_constraint(cons)
                cons_clean, _ = remove_unrelated_context(cons_simp, ignore_vars=set())
                yield cons_clean


def check_type_errors(
    ctx: TypingContext,
    term: Term,
    expected_type: Type,
) -> Iterable[AeonError]:
    if not wellformed(ctx, expected_type):
        return [CoreWellformnessError(expected_type)]

    try:
        constraint = check(ctx, term, expected_type)
        match entailment(ctx, constraint):
            case True:
                return []
            case False:
                full_constraint = entailment_context(ctx, constraint)
                return [
                    LiquidTypeCheckingFailedRelation(ctx, term, expected_type, v)
                    for v in constraint_to_parts(full_constraint)
                ]
    except CoreTypeCheckingError as e:
        return [e]



from typing import Callable
from time import monotonic_ns

from aeon.core.terms import Term
from aeon.core.types import Type
from aeon.decorators.api import Metadata
from aeon.synthesis.api import Synthesizer
from aeon.synthesis.uis.api import SynthesisUI
from aeon.typechecking.context import TypingContext
from aeon.utils.name import Name
from aeon.sugar.parser import parse_expression
from aeon.core.terms import Hole
from aeon.sugar.lowering import lower_to_core
from aeon.synthesis.decorators import Goal

from ollama import generate


def get_elapsed_time(start_time) -> float:
    """The elapsed time since the start in seconds."""
    return (monotonic_ns() - start_time) * 0.000000001


def is_better(a: list[float], b: list[float] | None, minimize: list[bool]) -> bool:
    if b is None:
        return True
    wins = 0
    losses = 0
    for ai, bi, min in zip(a, b, minimize):
        if min:
            if ai <= bi:
                wins += 1
            else:
                losses += 1
        else:
            if ai >= bi:
                wins += 1
            else:
                losses += 1
    return wins - losses > 0


class LLMSynthesizer(Synthesizer):
    def synthesize(
        self,
        ctx: TypingContext,
        type: Type,
        validate: Callable[[Term], bool],
        evaluate: Callable[[Term], list[float]],
        fun_name: Name,
        metadata: Metadata,
        budget: float = 60,
        ui: SynthesisUI = SynthesisUI(),
    ) -> Term:
        assert isinstance(ctx, TypingContext)
        assert isinstance(type, Type)

        var_description = ", ".join([f"{name} : {ty}" for (name, ty) in ctx.concrete_vars()])

        system_prompt = (
            "Please generate a candidate expression for the problem defined after the word PROBLEM."
            f"The candidate expression should be an expression of type {type}, and"
            "be written in the aeon programming language."
            "Aeon is a functional programming language, with a syntax very similar to Haskell, but with colons like ML."
            "Aeon has first-class refinement types, but unlike LiquidHaskell, those are not presented as comments, but rather directly in types."
            "Present the expression directly, with no explanation or code around it."
            "Presente the expression as you would enter it in an interpreter, without top-level declarations or type annotations."
            f"In the expression, you can use the following variables: {var_description}"
            "\n"
            "\nPROBLEM:"
        )
        core_term: Term = Hole(Name("sorry", -1))
        best_quality = None

        current_metadata = metadata.get(fun_name, {})
        goals: list[Goal] = current_metadata.get("goals", [])
        minimize_list = [goal.minimize for goal in goals for _ in range(goal.length)]
        prompt = current_metadata.get("prompt", "Any program")

        start_time = monotonic_ns()
        temperature = 0.0
        while get_elapsed_time(start_time) <= budget:
            result = generate(
                model="qwen2.5:32b", prompt=f"{system_prompt}\n{prompt}```", options={"temperature": temperature}
            )
            r = result.response
            try:
                tterm = parse_expression(f"({r})")
                core_tterm = lower_to_core(tterm)

                if validate(core_tterm):
                    quality = evaluate(core_tterm)
                    if len(quality) == 0:
                        return core_tterm
                    time = get_elapsed_time(start_time)
                    is_best = is_better(quality, best_quality, minimize_list)
                    ui.register(core_tterm, quality, time, is_best)
                    if is_best:
                        core_term = core_tterm
                        best_quality = quality
                else:
                    time = get_elapsed_time(start_time)
                    ui.register(core_tterm, None, time, False)

            except Exception:
                temperature += 0.1
        return core_term
